from typing import TYPE_CHECKING, Optional

from ._tools import _convert_date_time
from .._data._data_provider_layer import DataProviderLayer
from .._data._data_type import DataType
from ..._tools import validate_types, attributes_arg_parser, try_copy_to_list

if TYPE_CHECKING:
    from ..._types import OptStrStrs


class Definition(DataProviderLayer):
    """
    Definition object.

    Parameters
    __________
        name : str, optional
            Filter results by Bucket name. (Partial match).
        created_since : str, optional
            Filter results by when the Bucket was created.
            The date/time format is YYYY-MM-DDTHH:mm:ssZ.(e.g. 2020-01-21T04:19:01Z).
        modified_since : str, optional
            Filter results by when the Bucket was last modified.
            The date/time format is YYYY-MM-DDTHH:mm:ssZ.(e.g. 2020-01-21T04:19:01Z).
        available_from : str, optional
            Filter results by when the Bucket was made available.
            The date/time format is YYYY-MM-DDTHH:mm:ssZ.(e.g. 2020-01-21T04:19:01Z).
        available_to : str, optional
            Filter results by when the Bucket was made available.
            The date/time format is YYYY-MM-DDTHH:mm:ssZ.(e.g. 2020-01-21T04:19:01Z).
        attributes : list of str, optional
            List of publisher-defined bucket attributes.
        page_size : int, optional
            Number of buckets returned.
            By default 25.
        skip_token : str, optional
            Skip token is only used if a previous operation returned a partial result.
            If a previous response contains a nextLink element, the value of the nextLink
            element will include a skip token parameter that specifies a starting point to use
            for subsequent calls.

    Methods
    -------
    get_data(session=session)
        Returns a response to the data platform
    get_data_async(session=None)
        Returns a response asynchronously to the data platform

    Examples
    --------
     >>> from refinitiv.data.delivery import cfs
     >>> definition = cfs.buckets.Definition()
     >>> buckets = definition.get_data()

     Using get_data_async
     >>> import asyncio
     >>> task = definition.get_data_async()
     >>> response = asyncio.run(task)
    """

    def __init__(
        self,
        name: Optional[str] = None,
        created_since: Optional[str] = None,
        modified_since: Optional[str] = None,
        available_from: Optional[str] = None,
        available_to: Optional[str] = None,
        attributes: "OptStrStrs" = None,
        page_size: int = 25,
        skip_token: Optional[str] = None,
    ):
        validate_types(page_size, [int], "page_size")
        created_since = _convert_date_time(created_since)
        modified_since = _convert_date_time(modified_since)
        available_from = _convert_date_time(available_from)
        available_to = _convert_date_time(available_to)
        if attributes:
            attributes = try_copy_to_list(attributes)
            attributes = attributes_arg_parser.get_list(attributes)
            attributes = ";".join(attributes)
        super().__init__(
            data_type=DataType.CFS_BUCKETS,
            name=name,
            created_since=created_since,
            modified_since=modified_since,
            available_from=available_from,
            available_to=available_to,
            attributes=attributes,
            page_size=page_size,
            skip_token=skip_token,
        )



__version__ = "1.0.129"


from ._butterfly_shift import *
from ._combined_shift import *
from ._curve_definition_pricing import *
from ._flattening_shift import *
from ._fx_forward_curve_definition import *
from ._fx_forward_curve_parameters import *
from ._fx_shift_scenario import *
from ._long_end_shift import *
from ._parallel_shift import *
from ._shift_definition import *
from ._short_end_shift import *
from ._time_bucket_shift import *
from ._twist_shift import *
from ._valuation_time import *



import json
from types import SimpleNamespace

import pandas as pd
import pytest
from pandas import DataFrame

from refinitiv.data._content_type import ContentType
from refinitiv.data.content._content_data import Data
from refinitiv.data.content._df_build_type import DFBuildType
from refinitiv.data.content._df_builder_factory import get_dfbuilder
from refinitiv.data.content._historical_data_provider import (
    HistoricalDataProvider,
    copy_fields,
)
from refinitiv.data.content._join_responses import join_responses
from refinitiv.data.content._universe_content_validator import (
    UniverseContentValidator,
)
from refinitiv.data.delivery._data._data_provider import (
    ResponseFactory,
    ParsedData,
    Response,
)
from tests.unit.conftest import StubResponse


@pytest.mark.parametrize(
    ("input_data", "expected_result", "expected_data"),
    [
        (
            {"content_data": {"data": "data"}},
            True,
            {
                "status": {},
                "raw_response": {},
                "content_data": {"data": "data"},
                "error_codes": [],
                "error_messages": [],
            },
        ),
        (
            {"content_data": None},
            False,
            {
                "status": {},
                "raw_response": {},
                "content_data": None,
                "error_codes": [1],
                "error_messages": ["Content data is None"],
            },
        ),
        (
            {"content_data": {"error": {"code": 500, "description": "error_desc"}}},
            False,
            {
                "status": {},
                "raw_response": {},
                "content_data": {"error": {"code": 500, "description": "error_desc"}},
                "error_codes": [500],
                "error_messages": ["error_desc"],
            },
        ),
        (
            {"content_data": {"error": {"message": "error_message"}}},
            False,
            {
                "status": {},
                "raw_response": {},
                "content_data": {"error": {"message": "error_message"}},
                "error_codes": [None],
                "error_messages": ["error_message"],
            },
        ),
        (
            {
                "content_data": {
                    "error": {
                        "message": "error_message",
                        "errors": ["message1", "message2"],
                    }
                }
            },
            False,
            {
                "status": {},
                "raw_response": {},
                "content_data": {
                    "error": {
                        "message": "error_message",
                        "errors": ["message1", "message2"],
                    }
                },
                "error_codes": [None],
                "error_messages": ["error_message:\nmessage1\nmessage2"],
            },
        ),
        (
            {
                "content_data": {
                    "universe": [
                        {
                            "Instrument": "invalid",
                            "Organization PermID": "Failed to resolve identifier(s).",
                        }
                    ]
                }
            },
            True,
            {
                "status": {},
                "raw_response": {},
                "content_data": {
                    "universe": [
                        {
                            "Instrument": "invalid",
                            "Organization PermID": "Failed to resolve identifier(s).",
                        }
                    ]
                },
                "error_codes": [],
                "error_messages": ["Failed to resolve identifiers ['invalid']"],
            },
        ),
    ],
)
def test_content_validator(input_data, expected_result, expected_data):
    # given
    content_validator = UniverseContentValidator()

    # when
    parsed_data = ParsedData({}, {}, **input_data)
    is_valid = content_validator.validate(parsed_data)

    # then
    assert is_valid == expected_result
    assert parsed_data == expected_data


def test_content_validator_unable_to_resolve_all_requested_identifiers():
    # given
    content_validator = UniverseContentValidator()
    universe = '{"url":{"params": {}}}'
    universe = json.loads(universe, object_hook=lambda d: SimpleNamespace(**d))
    universe_data = {"universe": "A,B,C"}
    universe.url.params = universe_data
    data = ParsedData(
        {},
        **{
            "content_data": {
                "error": {
                    "code": 400,
                    "description": "Unable to resolve all requested identifiers.",
                }
            },
            "raw_response": universe,
        },
    )

    # when
    is_valid = content_validator.validate(data)

    # then
    assert not is_valid
    assert data.first_error_code == 400
    assert (
        data.first_error_message
        == f"Unable to resolve all requested identifiers. Requested items: {universe_data['universe'].split(',')}"
    )


@pytest.mark.parametrize(
    ("input_value", "expected_result"),
    [
        (
            {
                "raw": {
                    "headers": [{"name": "name", "title": "title"}],
                    "data": [[1], [2]],
                },
                "use_field_names_in_headers": True,
                "dfbuilder": get_dfbuilder(
                    ContentType.DATA_GRID_RDP, DFBuildType.INDEX
                ),
            },
            (["name"], 2),
        ),
        (
            {
                "raw": {
                    "headers": [{"name": "name", "title": "title"}],
                    "data": [[1], [2]],
                },
                "use_field_names_in_headers": False,
                "dfbuilder": get_dfbuilder(
                    ContentType.DATA_GRID_RDP, DFBuildType.INDEX
                ),
            },
            (["title"], 2),
        ),
        (
            {
                "raw": {
                    "headers": [{"name": "name", "title": "title"}],
                    "data": [[1], [2]],
                },
                "use_field_names_in_headers": False,
                "dfbuilder": get_dfbuilder(
                    ContentType.DATA_GRID_RDP, DFBuildType.INDEX
                ),
            },
            (["title"], 2),
        ),
    ],
)
def test_universe_data(input_value, expected_result):
    # given
    data_class = Data(**input_value)

    # when
    df = data_class.df

    # then
    # header name
    assert list(df) == expected_result[0]
    # size
    assert df.size == expected_result[1]


@pytest.mark.parametrize(
    ("input_value", "expected_result"),
    [
        (
            {
                "raw": {},
                "use_field_names_in_headers": False,
                "dfbuilder": get_dfbuilder(
                    ContentType.DATA_GRID_RDP, DFBuildType.INDEX
                ),
            },
            DataFrame(),
        ),
        (
            {
                "raw": {},
                "use_field_names_in_headers": True,
                "dfbuilder": get_dfbuilder(
                    ContentType.DATA_GRID_RDP, DFBuildType.INDEX
                ),
            },
            DataFrame(),
        ),
    ],
)
def test_universe_data_none(input_value, expected_result):
    # given
    data = Data(**input_value)

    # when
    df = data.df

    # then
    diff = df.compare(expected_result)
    assert diff.empty is True


@pytest.mark.parametrize(
    "input_value",
    [
        {
            # no headers
            "raw": {"data": []},
            "use_field_names_in_headers": True,
            "dfbuilder": get_dfbuilder(ContentType.DATA_GRID_RDP, DFBuildType.INDEX),
        },
        {
            # no data
            "raw": {"headers": []},
            "use_field_names_in_headers": True,
            "dfbuilder": get_dfbuilder(ContentType.DATA_GRID_RDP, DFBuildType.INDEX),
        },
    ],
)
def test_universe_data_empty_dataframe(input_value):
    # given
    data_class = Data(**input_value)

    # when
    df = data_class.df

    # then
    assert df.empty


def test_universe_response_factory_create_success():
    # given
    data = ParsedData({}, {}, **{"content_data": {"test": ""}})
    response_factory = ResponseFactory()

    # when
    result = response_factory.create_success(data)

    # then
    assert hasattr(result, "data")
    assert result.data.raw == data.content_data


def test_universe_response_factory_create_fail():
    # given
    data = ParsedData({}, {}, {"content_data": {"test": ""}})
    response_factory = ResponseFactory()

    # when
    result = response_factory.create_fail(data)

    # then
    assert hasattr(result, "data")
    assert result.data.raw == data.content_data


@pytest.mark.parametrize(
    "input_responses, expected_raise",
    [
        ([(1, "")], AttributeError),
        ([("1", "")], AttributeError),
        ([("1", [])], AttributeError),
    ],
)
def test_historical_data_provider_join_responses_not_valid_response(
    input_responses, expected_raise
):
    historical_data_provider = HistoricalDataProvider()
    historical_data_provider._get_axis_name = lambda **kwargs: ""
    # when
    with pytest.raises(expected_raise):
        historical_data_provider._join_responses(
            input_responses,
            data=[],
        )


@pytest.mark.parametrize(
    "input_value, expected_value",
    [
        ({"fields": None}, []),
        ({"fields": []}, []),
    ],
)
def test_success_without_valid_get_fields(input_value, expected_value):
    # when
    result = copy_fields(**input_value)

    # then
    assert isinstance(result, list)
    assert len(result) == 0


def test_success_with_valid_input_arg_get_fields():
    # given
    expected_value = ["IBA"]
    input_value = {"fields": expected_value}

    # when
    result = copy_fields(**input_value)

    # then
    assert isinstance(result, list)
    assert result == expected_value
    assert result is not expected_value


@pytest.mark.parametrize(
    "input_value, expected_raise",
    [
        ({"fields": ()}, AttributeError),
    ],
)
def test_failed_get_fields(input_value, expected_raise):
    # when
    with pytest.raises(expected_raise):
        copy_fields(**input_value)


def test__join_responses():
    # given
    http_statuses = [
        {"code": 200, "message": "status1"},
        {"code": 300, "message": "status2"},
    ]

    http_headers = ["header1", "header2"]
    request_messages = [0, 1]
    http_responses = [0, 1]

    response1 = StubResponse()
    response2 = StubResponse()

    response1.is_success = True
    response2.is_success = True

    response1.http_status = http_statuses[0]
    response2.http_status = http_statuses[1]

    response1.http_headers = http_headers[0]
    response2.http_headers = http_headers[1]

    response1.request_message = request_messages[0]
    response2.request_message = request_messages[1]

    response1.http_response = http_responses[0]
    response2.http_response = http_responses[1]
    response1.errors = []
    response2.errors = []

    response1.data = Data({}, pd.DataFrame({}))
    response2.data = Data({}, pd.DataFrame({}))

    responses = [response1, response2]

    # when
    result = join_responses(responses)

    # then
    assert isinstance(result, Response)
    assert result.http_response == http_responses
    assert result.http_headers == http_headers
    assert result.http_status == http_statuses
    assert result.request_message == request_messages



from typing import TYPE_CHECKING

from .._content_data import Data
from .._content_provider_layer import ContentUsageLoggerMixin
from ..._content_type import ContentType
from ...delivery._data._data_provider import DataProviderLayer, BaseResponse

if TYPE_CHECKING:
    from ..._types import ExtendedParams


class Definition(
    ContentUsageLoggerMixin[BaseResponse[Data]], DataProviderLayer[BaseResponse[Data]]
):
    """
    This class describe parameters to retrieve data for search custom instrument

    Parameters
    ----------
    access : str
        The search based on relationship to the custom instrument, for now only "owner" is supported. Can be omitted, default value is "owner"
    extended_params : dict, optional
        If necessary other parameters

    Examples
    --------
    >>> from refinitiv.data.content.custom_instruments import search
    >>> definition_search = search.Definition("VOD.L")
    >>> response = definition_search.get_data()
    """

    _USAGE_CLS_NAME = "CustomInstruments.SearchDefinition"

    def __init__(
        self,
        access: str = "owner",
        extended_params: "ExtendedParams" = None,
    ):
        super().__init__(
            data_type=ContentType.CUSTOM_INSTRUMENTS_SEARCH,
            access=access,
            extended_params=extended_params,
        )



# coding: utf8
from ..._instrument_definition import ObjectDefinition


class FxDoubleBarrierInfo(ObjectDefinition):
    """ """

    def __init__(self, in_or_out=None, level=None, rebate_amount=None):
        super().__init__()
        self.in_or_out = in_or_out
        self.level = level
        self.rebate_amount = rebate_amount

    @property
    def in_or_out(self):
        """
        In/Out property of the barrier option
        :return: string
        """
        return self._get_parameter("inOrOut")

    @in_or_out.setter
    def in_or_out(self, value):
        self._set_parameter("inOrOut", value)

    @property
    def level(self):
        """
        Barrier of the barrier option
        :return: double
        """
        return self._get_parameter("level")

    @level.setter
    def level(self, value):
        self._set_parameter("level", value)

    @property
    def rebate_amount(self):
        """
        Rebate of the barrier option
        :return: double
        """
        return self._get_parameter("rebateAmount")

    @rebate_amount.setter
    def rebate_amount(self, value):
        self._set_parameter("rebateAmount", value)



__all__ = (
    "Definition",
    "events",
    "summaries",
    "search",
    "manage",
    "Intervals",
    "CustomInstrumentTypes",
    "VolumeBasedRolloverMethod",
    "DayBasedRolloverMethod",
    "SpreadAdjustmentMethod"
)

from . import events, summaries, search, manage
from ._enums import CustomInstrumentTypes, VolumeBasedRolloverMethod, DayBasedRolloverMethod, SpreadAdjustmentMethod
from ._definition import Definition
from .._intervals import Intervals



import refinitiv.data.content.ipa.financial_contracts as rdf
from tests.integration.conftest import is_open
from tests.integration.constants_list import HttpStatusCode


def swap_definition_01():
    definition = rdf.swap.Definition(
        instrument_tag="user-defined GBP IRS",
        start_date="2019-05-21T00:00:00Z",
        tenor="10Y",
        legs=[
            rdf.swap.LegDefinition(
                direction=rdf.swap.Direction.PAID,
                notional_amount=10000000,
                notional_ccy="GBP",
                interest_type=rdf.swap.InterestType.FIXED,
                interest_payment_frequency=rdf.swap.Frequency.ANNUAL,
                interest_calculation_method=rdf.swap.DayCountBasis.DCB_30_360,
                payment_business_day_convention=rdf.swap.BusinessDayConvention.MODIFIED_FOLLOWING,
                payment_roll_convention=rdf.swap.DateRollingConvention.SAME,
                payment_business_days="UKG",
                amortization_schedule=[
                    rdf.swap.AmortizationItem(
                        remaining_notional=200000,
                        amortization_frequency=rdf.swap.AmortizationFrequency.EVERY_COUPON,
                        amortization_type=rdf.swap.AmortizationType.LINEAR,
                    )
                ],
            ),
            rdf.swap.LegDefinition(
                direction=rdf.swap.Direction.RECEIVED,
                notional_amount=10000000,
                notional_ccy="GBP",
                interest_type=rdf.swap.InterestType.FLOAT,
                interest_payment_frequency=rdf.swap.Frequency.SEMI_ANNUAL,
                index_reset_frequency=rdf.swap.Frequency.SEMI_ANNUAL,
                interest_calculation_method=rdf.swap.DayCountBasis.DCB_ACTUAL_360,
                payment_business_day_convention=rdf.swap.BusinessDayConvention.MODIFIED_FOLLOWING,
                payment_roll_convention=rdf.swap.DateRollingConvention.SAME,
                payment_business_days="UKG",
                spread_bp=20,
                index_name="SONIA",
                index_tenor="ON",
                index_reset_type=rdf.swap.IndexResetType.IN_ARREARS,
                amortization_schedule=[
                    rdf.swap.AmortizationItem(
                        remaining_notional=200000,
                        amortization_frequency=rdf.swap.AmortizationFrequency.EVERY2ND_COUPON,
                        amortization_type=rdf.swap.AmortizationType.LINEAR,
                    )
                ],
            ),
        ],
        pricing_parameters=rdf.swap.PricingParameters(discounting_tenor="ON"),
        fields=[
            "InstrumentTag",
            "InstrumentDescription",
            "FixedRatePercent",
            "MarketValueInDealCcy",
            "PV01Bp",
            "DiscountCurveName",
            "ForwardCurveName",
            "CashFlows",
            "ErrorMessage",
        ],
    )

    return definition


def swap_definition_02():
    definition = rdf.swap.Definition(
        instrument_tag="NOK_AB6O 5Y swap",
        template="NOK_AB6O",
        tenor="5Y",
        pricing_parameters=rdf.swap.PricingParameters(
            valuation_date="2018-01-10T00:00:00Z", report_ccy="JPY"
        ),
    )

    return definition


def swap_definition_03():
    definition = rdf.swap.Definition(
        tenor="2Y",
        legs=[
            rdf.swap.LegDefinition(
                interest_type=rdf.swap.InterestType.FLOAT,
                interest_payment_frequency=rdf.swap.Frequency.QUARTERLY,
                direction=rdf.swap.Direction.PAID,
                notional_amount=1,
                notional_ccy="EUR",
            ),
            rdf.swap.LegDefinition(
                index_tenor="5Y",
                interest_type=rdf.swap.InterestType.FLOAT,
                interest_payment_frequency=rdf.swap.Frequency.QUARTERLY,
                direction=rdf.swap.Direction.RECEIVED,
                notional_ccy="EUR",
            ),
        ],
        pricing_parameters=rdf.swap.PricingParameters(
            index_convexity_adjustment_integration_method=rdf.swap.IndexConvexityAdjustmentIntegrationMethod.RIEMANN_SUM,
            index_convexity_adjustment_method=rdf.swap.IndexConvexityAdjustmentMethod.BLACK_SCHOLES,
            valuation_date="2020-06-01",
        ),
        fields=[
            "InstrumentDescription",
            "Structure",
            "SettlementCcy",
            "InstrumentTag",
            "LegTag",
            "ValuationDate",
            "Calendar",
            "IndexRic",
            "IndexTenor",
            "EndDate",
            "StartDate",
            "Tenor",
            "InterestType",
            "NotionalCcy",
            "NotionalAmount",
            "MarketValueInDealCcy",
            "MarketValueInLegCcy",
            "CleanMarketValueInDealCcy",
            "CleanMarketValueInLegCcy",
            "AccruedAmountInDealCcy",
            "MarketValueInReportCcy",
            "CleanMarketValueInReportCcy",
            "AccruedAmountInReportCcy",
            "CleanPricePercent",
            "DirtyPricePercent",
            "AccruedPercent",
            "FixedRatePercent",
            "FixedRate",
            "SpreadBp",
            "ModifiedDuration",
            "Duration",
            "AnnuityAmountInDealCcy",
            "AnnuityAmountInReportCcy",
            "AnnuityBp",
            "DV01AmountInDealCcy",
            "DV01AmountInReportCcy",
            "DV01Bp",
            "PV01AmountInDealCcy",
            "PV01AmountInReportCcy",
            "PV01Bp",
            "DiscountCurveName",
            "ForwardCurveName",
            "ErrorMessage",
        ],
    )
    return definition


def check_http_status_is_success_and_df_value_not_empty(response):
    status = response.http_status
    df = response.data.df
    assert status["http_status_code"] == HttpStatusCode.TWO_HUNDRED, (
        f"Actual status code is {status['http_status_code']}, "
        f"Error: {response.http_status['error']}"
    )

    assert response.data.raw is not None, f"raw is {response.data.raw}"
    assert df is not None, f"DataFrame is {df}"
    assert not df.empty, f"DataFrame is empty: {df}"
    assert not bool(df.ErrorMessage[0]), f"{df.ErrorMessage[0]}"
    assert df.MarketValueInDealCcy is not None, f"{df.MarketValueInDealCcy}"


def check_stream_state_and_df_from_stream(stream):
    df = stream.get_snapshot()

    assert is_open(stream), f"Stream is not open"
    assert df is not None, f"DataFrame is {df}"
    assert not df.empty, f"DataFrame is empty: {df}"
    assert df.MarketValueInDealCcy is not None, f"{df.MarketValueInDealCcy}"
    assert not bool(df.ErrorMessage[0]), f"{df.ErrorMessage[0]}"

    stream.close()


swap_universe = {
    "instrumentType": "Swap",
    "instrumentDefinition": {
        "InstrumentTag": "my tag",
        "tenor": "2Y",
        "legs": [
            {
                "direction": "Paid",
                "interestType": "Float",
                "notionalCcy": "EUR",
                "notionalAmount": 1,
                "interestPaymentFrequency": "Quarterly",
            },
            {
                "direction": "Received",
                "interestType": "Float",
                "notionalCcy": "EUR",
                "indexTenor": "5Y",
                "interestPaymentFrequency": "Quarterly",
            },
        ],
    },
    "pricingParameters": {
        "indexConvexityAdjustmentIntegrationMethod": "RiemannSum",
        "indexConvexityAdjustmentMethod": "BlackScholes",
        "valuationDate": "2020-06-01",
    },
}


from ._image import Image, ResizedImage
from ..._content_data_provider import ContentDataProvider
from ...._tools import ParamItem
from ...._tools import extend_params
from ....delivery._data._data_provider import RequestFactory
from ..._content_response_factory import ContentResponseFactory
from ...._core.session import Session
from ...._tools import ParamItem, get_correct_filename
from ....delivery._data._data_provider import (
    RequestFactory,
)
from ....delivery._data._endpoint_data import EndpointData
from ....delivery._data._parsed_data import ParsedData
from ....delivery._data._response_factory import ResponseFactory
from ....delivery._data._validators import (
    ValidatorContainer,
    ContentValidator,
    ContentTypeValidator,
)


class ImageData(EndpointData):
    @property
    def image(self) -> "Image":
        if "image" in self.raw:
            return ResizedImage(self.raw)
        return Image(self.raw)


query_params = [
    ParamItem("width"),
    ParamItem("height"),
]


class ImagesRequestFactory(RequestFactory):
    def extend_query_parameters(self, query_parameters, extended_params=None):
        return extend_params(query_parameters, extended_params)

    @property
    def query_params_config(self):
        return query_params

    def get_header_parameters(
        self, session=None, header_parameters=None, width=None, height=None, **kwargs
    ):
        if width or height:
            return {"accept": "image/jpeg"}
        return {"accept": "application/json"}

    def get_path_parameters(self, session=None, *, image_id=None, **kwargs):
        return {"imageId": image_id}

    def get_url(self, *args, **kwargs):
        return f"{super().get_url(*args, **kwargs)}/{{imageId}}"


def replace_from_end(s: str, old: str, new: str, count: int) -> str:
    return s[::-1].replace(old, new, count)[::-1]


def get_extension_by_image_content_type(content_type: str) -> str:
    # content_type: 'image/svg+xml'
    extension = content_type.split("/")[1]
    # extension -> 'svg+xml'
    if "+" in extension:
        extension = extension.split("+")[0]
        # extension -> 'svg'
    return extension


def get_image_filename(image_id: str, content_type: str) -> str:
    extension = get_extension_by_image_content_type(content_type)
    filename = f"{image_id}.{extension}"
    return get_correct_filename(filename, "_")


class ImagesResponseFactory(ResponseFactory):
    def create_data_success(
        self,
        parsed_data: "ParsedData",
        image_id=None,
        **kwargs,
    ):
        headers = parsed_data.raw_response.headers
        content_type = dict(headers).get("content-type")
        if content_type.startswith("image/"):
            image_data = {
                "image": parsed_data.raw_response.content,
                "filename": get_image_filename(image_id, content_type),
            }
        else:
            image_data = self.get_raw(parsed_data)

        return self.data_class(image_data, **kwargs)


class ImagesContentValidator(ContentValidator):
    @property
    def validators(self):
        return [self.content_data_is_not_none]


news_images_data_provider = ContentDataProvider(
    request=ImagesRequestFactory(),
    response=ImagesResponseFactory(data_class=ImageData),
    validator=ValidatorContainer(
        content_validator=ImagesContentValidator(),
        content_type_validator=ContentTypeValidator({"application/json", "image/jpeg"}),
    ),
)



import refinitiv.data as rd

IPA_ITEMS = [
    rd.content.ipa.curves.forward_curves.Definition,
    rd.content.ipa.dates_and_calendars.add_periods.Definition,
    rd.content.ipa.dates_and_calendars.count_periods.Definition,
    rd.content.ipa.dates_and_calendars.date_schedule.Definition,
    rd.content.ipa.dates_and_calendars.holidays.Definition,
    rd.content.ipa.dates_and_calendars.is_working_day.Definition,
    rd.content.ipa.financial_contracts.bond.Definition,
    rd.content.ipa.financial_contracts.cap_floor.Definition,
    rd.content.ipa.financial_contracts.cds.Definition,
    rd.content.ipa.financial_contracts.cross.Definition,
    rd.content.ipa.financial_contracts.option.Definition,
    rd.content.ipa.financial_contracts.repo.Definition,
    rd.content.ipa.financial_contracts.swap.Definition,
    rd.content.ipa.financial_contracts.swaption.Definition,
    rd.content.ipa.financial_contracts.term_deposit.Definition,
]


def test_ipa_creation_without_arguments():
    for definition in IPA_ITEMS:
        assert definition()



import pytest
from pandas.core.dtypes.common import is_datetime64tz_dtype, is_datetime64_any_dtype

from refinitiv.data.content._historical_content_validator import (
    HistoricalContentValidator,
)
from refinitiv.data.content._historical_df_builder import HistoricalBuilder
from refinitiv.data.content._historical_response_factory import (
    HistoricalResponseFactory,
)
from refinitiv.data.content._intervals import Intervals
from refinitiv.data.content.historical_pricing import (
    EventTypes,
    Adjustments,
    MarketSession,
)
from refinitiv.data.content.historical_pricing._historical_pricing_request_factory import (
    HistoricalPricingSummariesRequestFactory,
    HistoricalPricingEventsRequestFactory,
    HistoricalPricingRequestFactory,
)

from refinitiv.data.delivery._data._data_provider import ParsedData
from refinitiv.data.delivery._data._endpoint_data import Error
from tests.unit.conftest import StubSession


def test_historical_pricing_summaries_request_factory_get_query_parameters():
    # given
    expected_value = [("interval", "P1D")]
    request_factory = HistoricalPricingSummariesRequestFactory()

    # when
    testing_value = request_factory.get_query_parameters(interval=Intervals.DAILY)

    # then
    assert testing_value == expected_value


def test_historical_pricing_summaries_request_factory_get_query_parameters_without_pass_argument():
    # given
    expected_value = []
    request_factory = HistoricalPricingSummariesRequestFactory()

    # when
    testing_value = request_factory.get_query_parameters()

    # then
    assert testing_value == expected_value


def test_historical_pricing_events_request_factory_get_query_parameters():
    # given
    expected_value = [("eventTypes", "quote")]
    request_factory = HistoricalPricingEventsRequestFactory()

    # when
    testing_value = request_factory.get_query_parameters(event_types=EventTypes.QUOTE)

    # then
    assert testing_value == expected_value


def test_historical_pricing_events_request_factory_get_query_parameters_without_pass_argument():
    # given
    expected_value = []
    request_factory = HistoricalPricingEventsRequestFactory()

    # when
    testing_value = request_factory.get_query_parameters()

    # then
    assert testing_value == expected_value


def test_historical_pricing_request_factory_get_url():
    # given
    input_value = "url"
    expected_value = "url/{universe}"
    request_factory = HistoricalPricingRequestFactory()

    # when
    testing_value = request_factory.get_url({}, input_value)

    # then
    assert testing_value == expected_value


def test_historical_pricing_request_factory_get_path_parameters():
    # given
    input_value = "universe"
    expected_value = {"universe": "universe"}
    request_factory = HistoricalPricingRequestFactory()

    # when
    session = StubSession()
    testing_value = request_factory.get_path_parameters(session, universe=input_value)

    # then
    assert testing_value == expected_value


def test_historical_pricing_summaries_request_factory_get_query_parameters_all():
    # given
    expected_value = [
        ("start", "2019-01-19T09:00:00.000000000Z"),
        ("end", "2019-01-24T17:00:00.000000000Z"),
        ("adjustments", "RPO"),
        ("sessions", "post"),
        ("count", 2),
        ("fields", "field_1,field_2,DATE"),
    ]
    request_factory = HistoricalPricingSummariesRequestFactory()

    # when
    testing_value = request_factory.get_query_parameters(
        start="2019-01-19 09:00:00",
        end="2019-01-24 17:00:00",
        adjustments=Adjustments.RPO,
        sessions=MarketSession.POST,
        count=2,
        fields=["field_1", "field_2"],
    )

    # then
    assert testing_value == expected_value


def test_historical_pricing_events_request_factory_get_query_parameters_all():
    # given
    expected_value = [
        ("start", "2019-01-19T09:00:00.000000000Z"),
        ("end", "2019-01-24T17:00:00.000000000Z"),
        ("adjustments", "RPO"),
        ("count", 2),
        ("fields", "field_1,field_2,DATE_TIME"),
    ]
    request_factory = HistoricalPricingEventsRequestFactory()

    # when
    testing_value = request_factory.get_query_parameters(
        start="2019-01-19 09:00:00",
        end="2019-01-24 17:00:00",
        adjustments=Adjustments.RPO,
        sessions=MarketSession.POST,
        count=2,
        fields=["field_1", "field_2"],
    )

    # then
    assert testing_value == expected_value


def test_historical_pricing_request_factory_extend_query_parameters():
    # given
    expected_value = [
        ("start", "2019-01-19T09:00:00.000000000Z"),
        ("end", "2019-01-24T17:00:00.000000000Z"),
        ("count", 2),
    ]
    request_factory = HistoricalPricingRequestFactory()

    # when
    testing_value = request_factory.extend_query_parameters(
        query_parameters=[],
        extended_params={
            "start": "2019-01-19T09:00:00",
            "end": "2019-01-24T17:00:00",
            "count": 2,
        },
    )

    # then
    assert testing_value == expected_value


def test_historical_pricing_request_factory_extend_body_parameters():
    # given
    request_factory = HistoricalPricingRequestFactory()

    # when
    testing_value = request_factory.extend_body_parameters(
        body_parameters={"a": 10}, extended_params={"count": 2}
    )

    # then
    assert testing_value is None


def test_historical_pricing_request_factory_get_query_parameters_raise_error_if_count_less_one():
    # given
    request_factory = HistoricalPricingSummariesRequestFactory()

    # then
    with pytest.raises(ValueError, match="Count minimum value is 1"):
        # when
        request_factory.get_query_parameters(count=-1)


def test_historical_pricing_request_factory_get_query_parameters_without_pass_arguments():
    # given
    expected_value = []
    request_factory = HistoricalPricingRequestFactory()

    # when
    testing_value = request_factory.get_query_parameters()

    # then
    assert testing_value == expected_value


@pytest.mark.parametrize(
    "input_content_data",
    [
        [{"status": {"code": "Error"}}],
        [{"status": {"code": "UserRequestError"}}],
        [{"status": {"code": "Error"}, "universe": "IBM"}],
    ],
)
def test_historical_pricing_content_validator_validate_content_data_is_false(
    input_content_data,
):
    # given
    input_value = ParsedData({}, {}, **{"content_data": input_content_data})
    expected_value = False
    validator = HistoricalContentValidator()

    # when
    testing_value = validator.validate(input_value)

    # then
    assert testing_value == expected_value


@pytest.mark.parametrize(
    "input_content_data",
    [
        [{"status": {"code": ""}, "data": [{}]}],
        [{"status": {"code": ""}, "universe": "IBM", "data": [{}]}],
        [{"status": {"code": "UserRequestError"}, "universe": "IBM", "data": [{}]}],
    ],
)
def test_historical_pricing_content_validator_validate_content_data_is_true(
    input_content_data,
):
    # given
    input_value = ParsedData({}, {}, **{"content_data": input_content_data})
    expected_value = True
    validator = HistoricalContentValidator()

    # when
    testing_value = validator.validate(input_value)

    # then
    assert testing_value == expected_value


@pytest.mark.parametrize(
    "data",
    [
        (
            {
                "status": {},
                "raw_response": {},
                "content_data": [{"universe": {"ric": "IBM.N"}}],
            }
        )
    ],
)
def test_historical_pricing_response_factory_create_success(data):
    response_factory = HistoricalResponseFactory()
    parsed_data = ParsedData(**data)
    response = response_factory.create_success(parsed_data)

    assert response is not None
    assert len(response.errors) == 0


@pytest.mark.parametrize(
    "data,kwargs,error_code,error_message",
    [
        (
            {
                "content_data": [
                    {
                        "universe": {"ric": "VOD.L"},
                        "status": {
                            "code": "TS.Intraday.UserRequestError.90006",
                            "message": "The universe does not support the "
                            "following fields: [ASKee].",
                        },
                    }
                ]
            },
            {"fields": ["ASK", "ASKeee", "DATE_TIME"]},
            "TS.Intraday.UserRequestError.90006",
            "The universe does not support the following fields: [ASKee]. "
            "Requested ric: VOD.L",
        ),
        (
            {
                "content_data": [
                    {
                        "universe": {"ric": "LSEG.L"},
                        "status": {
                            "code": "TS.Intraday.Warning.95004",
                            "message": "Trades interleaving with corrections is "
                            "currently not supported. "
                            "Corrections will not be returned.",
                        },
                    }
                ]
            },
            {},
            "TS.Intraday.Warning.95004",
            "LSEG.L - Trades interleaving with corrections is currently not supported. "
            "Corrections will not be returned.",
        ),
    ],
)
def test_historical_pricing_response_factory_create_success_with_errs(
    data, kwargs, error_code, error_message
):
    response_factory = HistoricalResponseFactory()
    parsed_data = ParsedData({}, {}, **data)
    result = response_factory.create_success(parsed_data, **kwargs)
    assert isinstance(result.errors[0], Error)
    assert result.errors[0][0] == error_code
    assert result.errors[0][1] == error_message


@pytest.mark.parametrize(
    "data,kwargs,error_code,error_message",
    [
        (
            {
                "content_data": [{"universe": {"ric": "VOD.L"}}],
                "error_codes": "TS.Intraday.UserRequestError.90006",
                "error_messages": "The universe does not support the following fields: [ASKeee].",
            },
            {"fields": ["ASKeee", "DATE_TIME"]},
            "TS.Intraday.UserRequestError.90006",
            "The universe does not support the following fields: [ASKeee]. Requested ric: VOD.L",
        ),
        (
            {
                "content_data": [{"universe": {"ric": "VOD.LLL"}}],
                "error_codes": "TS.Intraday.UserRequestError.90001",
                "error_messages": "The universe is not found.",
            },
            {"fields": ["ASK"]},
            "TS.Intraday.UserRequestError.90001",
            "VOD.LLL - The universe is not found",
        ),
        (
            {"error_codes": 400, "error_messages": "Request Validation Error"},
            {"fields": ["TR.Volume", "DATE_TIME"], "universe": "IBM.N"},
            400,
            "Request Validation Error. Requested ric: IBM.N. Requested fields: ['TR.Volume', 'DATE_TIME']",
        ),
    ],
)
def test_historical_pricing_response_factory_create_fail(
    data, kwargs, error_code, error_message
):
    response_factory = HistoricalResponseFactory()
    parsed_data = ParsedData({}, {}, **data)
    result = response_factory.create_fail(parsed_data, **kwargs)
    assert hasattr(result, "errors")
    assert isinstance(result.errors[0], Error)
    assert result.errors[0][0] == error_code
    assert result.errors[0][1] == error_message


@pytest.mark.parametrize(
    "input_header_name,input_data",
    [
        ("DATE", [["2022-04-26", 130], ["2022-04-25", 129]]),
        (
            "DATE_TIME",
            [
                ["2021-07-27T16:30:00.000000000Z", 130],
                ["2021-07-27T15:50:00.000000000Z", 129],
            ],
        ),
    ],
)
def test_historical_pricing_build_df_index_type(input_header_name, input_data):
    # given
    input_raw = {
        "headers": [
            {"name": input_header_name, "type": "string"},
            {"name": "TRDPRC_1", "type": "number", "decimalChar": "."},
        ],
        "data": input_data,
        "universe": {"ric": ""},
    }

    df_builder = HistoricalBuilder()

    # when
    result_df = df_builder.build_one(input_raw, [], "")

    # then
    assert is_datetime64_any_dtype(result_df.index)
    assert not is_datetime64tz_dtype(result_df.index)



import refinitiv.data as rd
import tests.integration.conftest as global_conftest
from tests.integration.helpers import compare_list


def open_platform_session_for_omm_stream():
    return rd.open_session(config_name=global_conftest.conf, name="platform.my-session")


def open_deployed_session_for_omm_stream():
    return rd.open_session(
        config_name=global_conftest.conf, name="platform.deployed-cipsnylab2"
    )


def on_ack(ack_msg, stream, event_list: list = None):
    if event_list is not None:
        event_list.append(ack_msg)
    print(ack_msg)


def on_error(error_msg, stream, event_list: list = None):
    if event_list is not None:
        event_list.append(error_msg)
    print(error_msg)


def check_triggered_events(triggered_events, expected_events):
    for i, event_type in enumerate(expected_events):
        event_list = list(triggered_events[i].keys())
        assert event_list[0] == event_type

    actual_event = []
    for event in triggered_events:
        for event_name, event_body in event.items():
            actual_event.append(event_name)
            if event_name == "Error":
                assert (
                    False
                ), f"Error {event_body['State']['Text']} is found in the stream"

    for event_type in expected_events:
        assert event_type in actual_event, f"Not found '{event_type}' in the stream"


def get_error_event_text_msg(triggered_events):
    for event in triggered_events:
        for event_name, event_body in event.items():
            if event_name == "Status":
                return event_body["State"]["Text"]


def check_stream_view(expected_view, triggered_events):
    if expected_view is not None:
        if isinstance(expected_view, str):
            expected_view = [expected_view]
        if expected_view != []:
            for events in triggered_events:
                for _, event in events.items():
                    received_view = list(event.get("Fields").keys())
                    compare_list(received_view, expected_view)


def check_stream_universe(expected_universe, triggered_events_list):
    for events in triggered_events_list:
        for _, event in events.items():
            assert expected_universe == event.get("Key").get("Name")


def check_extended_params(stream, extended_params, triggered_events):
    for events in triggered_events:
        for _, event in events.items():
            assert extended_params["View"] == stream._extended_params["View"]
            received_view = list(event.get("Fields").keys())
            compare_list(received_view, extended_params["View"])



import allure
import pytest

from refinitiv.data.content.ipa.curves import zc_curve_definitions
from tests.integration.content.ipa.curves.conftest import on_response
from tests.integration.content.ipa.curves.zc_curve_definition.conftest import (
    check_sc_curve_definitions_response,
    zc_curve_definition_01,
    zc_curve_definition_02,
    invalid_zc_curve_definition,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_dataframe_column_date_for_datetime_type,
)


@allure.suite("Content object - ZcCurveDefinitions")
@allure.feature("Content object - ZcCurveDefinitions")
@allure.severity(allure.severity_level.CRITICAL)
class TestZcCurveDefinitions:
    @allure.title(
        "Create a zc_curve_definitions object with valid definitions and get data"
    )
    @pytest.mark.caseid("39785336")
    @pytest.mark.smoke
    def test_create_zc_curve_definition_with_valid_definitions(self, open_session):
        response = zc_curve_definitions.Definitions(
            [zc_curve_definition_01(), zc_curve_definition_02()]
        ).get_data(
            on_response=lambda response, definition, session: on_response(
                response, definition, open_session
            )
        )

        check_sc_curve_definitions_response(response=response, expected_currency="EUR")
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title(
        "Create zc curve definitions object with invalid definition and get data"
    )
    @pytest.mark.caseid("39785337")
    def test_create_invalid_zc_curve_definitions(self, open_session):
        response = zc_curve_definitions.Definitions(
            invalid_zc_curve_definition()
        ).get_data()

        assert response.data.df.empty

    @allure.title("Create a zc curve definitions object async")
    @pytest.mark.caseid("39785338")
    async def test_create_zc_curve_definitions_async(self, open_session_async):
        response_01, response_02 = await get_async_response_from_definitions(
            zc_curve_definitions.Definitions(zc_curve_definition_01()),
            zc_curve_definitions.Definitions(invalid_zc_curve_definition()),
        )

        check_sc_curve_definitions_response(
            response=response_01, expected_currency="EUR"
        )
        assert response_02.data.df.empty



from typing import Union, Iterable, Callable

from .._core.session import Session
from .._content_type import ContentType
from ..content._universe_streams import _UniverseStreams
from ..content.pricing._stream_facade import PricingStream


class StreamingPrices(_UniverseStreams):
    def __init__(
        self,
        universe: Union[str, Iterable[str]],
        session: "Session" = None,
        fields: Union[str, list] = None,
        service: str = None,
        on_refresh: Callable = None,
        on_status: Callable = None,
        on_update: Callable = None,
        on_complete: Callable = None,
        extended_params: dict = None,
    ) -> None:
        super().__init__(
            content_type=ContentType.STREAMING_PRICING,
            item_facade_class=PricingStream,
            universe=universe,
            session=session,
            fields=fields,
            service=service,
            on_refresh=on_refresh,
            on_status=on_status,
            on_update=on_update,
            on_complete=on_complete,
            extended_params=extended_params,
        )



import asyncio
import time

import allure
import pytest

from refinitiv.data import open_pricing_stream
from tests.integration.access.conftest import (
    check_stream_opened,
    check_stream_closed,
    check_stream_fields,
    check_df_recording_duration,
    check_df_recording_frequency,
    check_column_names_is_exist_in_response_and_df_not_empty,
    check_history_for_expected_amount_of_bars,
    is_universe_exist,
)
from tests.integration.helpers import (
    check_if_dataframe_is_not_none,
)


@allure.suite("FinCoder layer")
@allure.feature("FinCoder - Stream Recorder")
@allure.severity(allure.severity_level.CRITICAL)
class TestPricingStreamRecorder:
    @allure.title("Stream recording ohlc data with several universes and fields")
    @pytest.mark.parametrize(
        "universe,fields,expected_fields,frequency,duration,ticks_per_bar",
        [
            (
                ["EUR=", "GBP="],
                ["BID", "ASK"],
                [
                    ("EUR=", "ASK", "open"),
                    ("EUR=", "ASK", "high"),
                    ("EUR=", "ASK", "low"),
                    ("EUR=", "ASK", "close"),
                    ("EUR=", "BID", "open"),
                    ("EUR=", "BID", "high"),
                    ("EUR=", "BID", "low"),
                    ("EUR=", "BID", "close"),
                    ("GBP=", "ASK", "open"),
                    ("GBP=", "ASK", "high"),
                    ("GBP=", "ASK", "low"),
                    ("GBP=", "ASK", "close"),
                    ("GBP=", "BID", "open"),
                    ("GBP=", "BID", "high"),
                    ("GBP=", "BID", "low"),
                    ("GBP=", "BID", "close"),
                    ("ticks count", "", ""),
                ],
                "3s",
                "9s",
                "1",
            ),
            (
                ["GBP="],
                ["BID", "ASK"],
                [
                    ("GBP=", "ASK", "open"),
                    ("GBP=", "ASK", "high"),
                    ("GBP=", "ASK", "low"),
                    ("GBP=", "ASK", "close"),
                    ("GBP=", "BID", "open"),
                    ("GBP=", "BID", "high"),
                    ("GBP=", "BID", "low"),
                    ("GBP=", "BID", "close"),
                    ("ticks count", "", ""),
                ],
                "tick",
                "13s",
                "3",
            ),
        ],
    )
    @pytest.mark.caseid("39028006")
    def test_stream_recording_ohlc_data_with_several_universes_and_fields(
        self,
        open_desktop_session,
        universe,
        fields,
        expected_fields,
        frequency,
        duration,
        ticks_per_bar,
    ):
        stream = open_pricing_stream(universe, fields)

        check_stream_opened(stream)
        stream.recorder.record(
            frequency=frequency, duration=duration, ticks_per_bar=ticks_per_bar
        )

        history_response = stream.recorder.get_history()
        stream.close()
        check_stream_closed(stream)

        check_stream_fields(stream, universe[0], fields)
        check_column_names_is_exist_in_response_and_df_not_empty(
            history_response, expected_fields
        )
        check_if_dataframe_is_not_none(history_response)
        if ticks_per_bar != "1":
            dataframes = stream.recorder._ohlc_builder.dataframes
            for i in range(len(dataframes) - 1):
                assert len(dataframes[i]) == int(ticks_per_bar)
            assert len(dataframes[-1]) <= int(ticks_per_bar)

    @allure.title(
        "Recording a pricing stream updates with closed stream and get ConnectionError"
    )
    @pytest.mark.parametrize(
        "universe,fields,frequency,duration",
        [
            (
                ["EUR=", "GBP="],
                ["BID", "ASK"],
                "3s",
                "9s",
            )
        ],
    )
    @pytest.mark.caseid("40101747")
    def test_record_pricing_stream_updates_with_closed_stream(
        self, open_desktop_session, universe, fields, frequency, duration
    ):
        with pytest.raises(ConnectionError) as error:
            stream = open_pricing_stream(universe, fields)
            stream.close()
            stream.recorder.record(frequency=frequency, duration=duration)
        assert str(error.value) == "Stream is closed. Cannot record."

    @allure.title("Check if recorded history is empty after delete method was applied")
    @pytest.mark.parametrize(
        "universe,fields,frequency,duration",
        [
            (
                ["EUR=", "GBP="],
                ["BID", "ASK"],
                "2s",
                "6s",
            )
        ],
    )
    @pytest.mark.caseid("40101749")
    def test_check_if_history_is_empty_after_record_history_was_deleted(
        self, open_desktop_session, universe, fields, frequency, duration
    ):
        stream = open_pricing_stream(universe=universe, fields=fields)
        stream.recorder.record(frequency=frequency, duration=duration)
        history_response_01 = stream.recorder.get_history()
        check_history_for_expected_amount_of_bars(
            stream, history_response_01, frequency, duration
        )
        stream.recorder.delete()
        assert stream.recorder._ohlc_builder.ohlc_df is None

        stream.recorder.record(frequency=frequency, duration=duration)
        history_response_02 = stream.recorder.get_history()
        check_history_for_expected_amount_of_bars(
            stream, history_response_02, frequency, duration
        )
        stream.recorder.delete()
        assert stream.recorder._ohlc_builder.ohlc_df is None

        stream.recorder.record(frequency=frequency, duration=duration)
        history_response_03 = stream.recorder.get_history()
        check_history_for_expected_amount_of_bars(
            stream, history_response_03, frequency, duration
        )
        stream.recorder.delete()
        assert stream.recorder._ohlc_builder.ohlc_df is None

    @allure.title(
        "Recording a pricing stream ohlc data with few recorders and defined duration"
    )
    @pytest.mark.parametrize(
        "universe,fields,frequency,duration",
        [
            (
                ["EUR=", "GBP=", "JPY="],
                ["BID", "ASK"],
                "2s",
                "7s",
            )
        ],
    )
    @pytest.mark.caseid("40113562")
    def test_recording_pricing_stream_ohlc_data_with_few_recorders(
        self, open_desktop_session, universe, fields, frequency, duration
    ):
        stream = open_pricing_stream(universe=universe, fields=fields)

        stream.recorder.record(frequency=frequency, duration=duration)
        history_response_01 = stream.recorder.get_history()
        check_history_for_expected_amount_of_bars(
            stream, history_response_01, frequency, duration
        )

        stream.recorder.record(frequency=frequency, duration=duration)
        history_response_02 = stream.recorder.get_history()
        check_history_for_expected_amount_of_bars(
            stream, history_response_02, frequency, duration, recorders_amount=2
        )

        stream.recorder.record(frequency=frequency, duration=duration)
        history_response_03 = stream.recorder.get_history()
        check_history_for_expected_amount_of_bars(
            stream, history_response_03, frequency, duration, recorders_amount=3
        )

    @allure.title("Recording a pricing stream ohlc data with callbacks")
    @pytest.mark.parametrize(
        "universe,fields,frequency,duration,time_to_sleep",
        [("GBP=", ["BID"], "3s", "15s", 0), ("GBP=", ["BID"], "3s", None, 5)],
    )
    @pytest.mark.caseid("39028007")
    def test_recording_pricing_stream_ohlc_data_with_callbacks(
        self,
        open_desktop_session,
        universe,
        fields,
        frequency,
        duration,
        time_to_sleep,
    ):
        callback_history = []
        stream = open_pricing_stream(universe, fields)
        start_time = time.time()
        stream.recorder.record(
            frequency=frequency,
            duration=duration,
            on_data=lambda history, recorder: callback_history.append(history),
        )
        if duration == None:
            duration = 15
            asyncio.run(asyncio.sleep(duration))
            stream.recorder.stop()
            history_01 = stream.recorder.get_history()
            time.sleep(time_to_sleep)
            history_02 = stream.recorder.get_history()
            assert (
                history_02.shape == history_01.shape
            ), f"Recorder wasn't stopped, recordered additional updates after stop() method"

        elapsed_time_secs = time.time() - start_time - time_to_sleep
        history_response = stream.recorder.get_history()
        stream.close()

        check_history_for_expected_amount_of_bars(
            stream, history_response, frequency, duration
        )
        check_df_recording_duration(elapsed_time_secs, duration)
        check_df_recording_frequency(history_response, frequency)
        assert len(callback_history) == len(
            history_response.values
        ), f"Amount of callback updates is not suited with recorded history"
        for history_update in callback_history:
            assert (
                history_update.values in history_response.values
            ), f"Found update row '{history_update}' is not present in recorded history"

    @allure.title("Recording a pricing updates data during defined duration")
    @pytest.mark.parametrize(
        "universe,fields,expected_fields,frequency,duration",
        [
            (
                ["EUR=", "GBP="],
                ["BID", "ASK", "CF_NETCHNG"],
                [
                    ("GBP=", "CF_NETCHNG"),
                    ("EUR=", "BID"),
                    ("EUR=", "ASK"),
                    ("GBP=", "BID"),
                    ("EUR=", "CF_NETCHNG"),
                    ("GBP=", "ASK"),
                ],
                "tick",
                "5s",
            ),
            (
                ["EUR=", "GBP="],
                [],
                [],
                "tick",
                "5s",
            ),
        ],
    )
    @pytest.mark.caseid("39028009")
    def test_stream_recording_all_updates_until_duration_ends(
        self,
        open_desktop_session,
        universe,
        fields,
        expected_fields,
        duration,
        frequency,
    ):
        stream = open_pricing_stream(universe, fields)
        stream.recorder.record(duration=duration)
        stream.recorder.stop()
        stream.close()
        check_stream_closed(stream)
        history_response = stream.recorder.get_history()

        check_stream_fields(stream, universe, fields)
        check_column_names_is_exist_in_response_and_df_not_empty(
            history_response, expected_fields
        )
        check_if_dataframe_is_not_none(history_response)
        assert (
            stream.recorder._ohlc_builder is None
        ), f"There is inconsistency in recorded data"

    @allure.title("Stream recording with mix of valid and invalid parameters")
    @pytest.mark.parametrize(
        "universes,fields,expected_fields,frequency,duration",
        [
            (
                ["GBP=", "INVALID"],
                ["BID", "ASK", "INVALID"],
                [
                    ("GBP=", "BID"),
                    ("GBP=", "ASK"),
                ],
                "tick",
                "7s",
            ),
            (
                ["GBP=", "INVALID"],
                ["BID", "ASK", "INVALID"],
                [
                    ("GBP=", "ASK", "open"),
                    ("GBP=", "ASK", "high"),
                    ("GBP=", "ASK", "low"),
                    ("GBP=", "ASK", "close"),
                    ("GBP=", "BID", "open"),
                    ("GBP=", "BID", "high"),
                    ("GBP=", "BID", "low"),
                    ("GBP=", "BID", "close"),
                    ("ticks count", "", ""),
                ],
                "3s",
                "7s",
            ),
        ],
    )
    @pytest.mark.caseid("39028012")
    def test_stream_recording_with_mix_of_valid_and_invalid_params(
        self,
        open_desktop_session,
        universes,
        fields,
        expected_fields,
        frequency,
        duration,
    ):
        stream = open_pricing_stream(universes, fields)
        stream.recorder.record(frequency=frequency, duration=duration)
        stream.close()
        history_response = stream.recorder.get_history()

        check_column_names_is_exist_in_response_and_df_not_empty(
            history_response, expected_fields
        )
        check_if_dataframe_is_not_none(history_response)
        assert not is_universe_exist(
            history_response, universes[1:]
        ), f"Expected invalid universe was met in response dataframe"
        assert is_universe_exist(
            history_response, universes[:-1]
        ), f"Expected valid universe was not met in response dataframe"



from typing import Union, Optional, Callable, TYPE_CHECKING

from ._type import contrib_type_enum_arg_parser
from ..event import StreamCxnEvent
from ..stream_connection import LOGIN_STREAM_ID
from ..._stream import _OMMStream
from ...._tools import cached_property, OrEvent

if TYPE_CHECKING:
    from ._type import ContribType
    from .._stream_factory import StreamDetails
    from ...._core.session import Session


class _OffStreamContrib(_OMMStream):
    def __init__(
        self,
        post_id: int,
        session: "Session",
        name: str,
        details: "StreamDetails",
        service: str,
        domain: str,
        on_ack: Optional[Callable] = None,
        on_error: Optional[Callable] = None,
    ) -> None:
        super().__init__(
            stream_id=post_id,
            session=session,
            name=name,
            details=details,
            domain=domain,
            service=service,
            on_error=on_error,
            on_ack=on_ack,
        )
        self._post_id = post_id

    @cached_property
    def _contributed(self) -> OrEvent:
        return OrEvent(self._error_event, self._ack_event)

    def get_next_post_id(self) -> int:
        return self.post_id

    def get_contrib_message(
        self, fields: dict, contrib_type: Union[str, "ContribType", None]
    ) -> dict:
        return {
            "Ack": True,
            "ID": LOGIN_STREAM_ID,
            "Message": {
                "Fields": fields,
                "ID": 0,
                "Type": contrib_type_enum_arg_parser.get_str(
                    contrib_type if contrib_type else "Update"
                ),
                "Domain": self.domain,
            },
            "PostID": self.post_id,
            "Type": "Post",
            "Key": {"Name": self.name, "Service": self.service},
            "Domain": self.domain,
        }

    def _do_open(self, *args, **kwargs):
        self._initialize_cxn()
        self._cxn.on(self._event.ack_by_id, self._on_stream_ack)
        self._cxn.on(self._event.error_by_id, self._on_stream_error)
        self._cxn.on(StreamCxnEvent.RECONNECTING, self.halt)

    def _do_close(self, *args, **kwargs):
        self._dispose()

    def _dispose(self):
        self._debug(f"{self._classname} disposing [d]")
        if self._cxn is not None:
            self._cxn.remove_listener(self._event.error_by_id, self._on_stream_error)
            self._cxn.remove_listener(self._event.ack_by_id, self._on_stream_ack)
            self._cxn.remove_listener(StreamCxnEvent.RECONNECTING, self.halt)
            self._release_cxn()
        self._contributed.set()
        self._debug(f"{self._classname} disposed [D]")



def test_cross_stream_has_correct_field_in_request_message(cross_stream):
    try:
        # when
        cross_stream._stream.universe["instrumentDefinition"]
    except Exception as e:
        assert False, str(e)

    else:
        assert True


def test_cross_stream_doesnot_have_incorrect_field_in_request_message(cross_stream):
    try:
        # when
        cross_stream._stream.universe["InstrumentDefinition"]
    except Exception as e:
        assert True, str(e)

    else:
        assert False



import refinitiv.data.content.ipa.financial_contracts as rdf


def bond_definition_fixed_rate():
    definition = rdf.bond.Definition(
        instrument_code="US10YT=RR",
        instrument_tag="TreasuryBond_10Y",
        fields=[
            "InstrumentCode",
            "BondType",
            "IssueDate",
            "MarketDataDate",
            "EndDate",
            "CouponRatePercent",
            "Accrued",
            "YieldPercent",
            "RedemptionDate",
            "ModifiedDuration",
            "Duration",
            "DV01Bp",
            "AverageLife",
            "Convexity",
            "ErrorMessage",
        ],
    )

    return definition


def bond_definition_floating_rate():
    definition = rdf.bond.Definition(
        instrument_code="61760LCZ6=",
        instrument_tag="TreasuryBond_10Y",
        pricing_parameters=rdf.bond.PricingParameters(market_data_date="2020-02-05"),
        fields=[
            "InstrumentCode",
            "BondType",
            "MarketDataDate",
            "ValuationDate",
            "IndexRic",
            "IborRatePercent",
            "Price",
            "AdjustedPrice",
            "NeutralPrice",
            "DiscountMarginBp",
            "SimpleMarginBp",
            "NeutralYieldPercent",
            "AccruedDays",
            "Accrued",
            "ErrorMessage",
        ],
    )

    return definition


def bond_definition_user_defined():
    definition = rdf.bond.Definition(
        issue_date="2002-02-28",
        end_date="2032-02-28",
        notional_ccy="USD",
        interest_payment_frequency="Annual",
        fixed_rate_percent=7,
        interest_calculation_method=rdf.bond.DayCountBasis.DCB_ACTUAL_ACTUAL,
        pricing_parameters=rdf.bond.PricingParameters(clean_price=122),
    )

    return definition


def invalid_bond_definition():
    definition = rdf.bond.Definition(
        instrument_code="INVALID_INSTRUMENT_CODE",
        instrument_tag="TreasuryBond_10Y",
        fields=["InstrumentCode", "IssueDate", "DirtyPrice", "DV01Bp", "ErrorMessage"],
    )

    return definition


bond_universe = {
    "instrumentType": "Bond",
    "instrumentDefinition": {
        "endDate": "2032-02-28",
        "notionalCcy": "USD",
        "fixedRatePercent": 9,
        "interestPaymentFrequency": "Annual",
        "interestCalculationMethod": "Dcb_Actual_Actual",
        "issueDate": "2002-02-28",
    },
    "pricingParameters": {"cleanPrice": 122},
}



from ._adjust_interest_to_payment_date import *
from ._american_monte_carlo_method import *
from ._amortization_frequency import *
from ._amortization_type import *
from ._average_type import *
from ._axis import *
from ._barrier_mode import *
from ._barrier_style import *
from ._barrier_type import *
from ._benchmark_yield_selection_mode import *
from ._binary_type import *
from ._business_day_convention import *
from ._buy_sell import *
from ._calibration_strategy import *
from ._call_put import *
from ._cds_convention import *
from ._credit_spread_type import *
from ._date_moving_convention import *
from ._date_rolling_convention import *
from ._dates_calendars_frequency import *
from ._day_count_basis import *
from ._day_of_week import *
from ._direction import *
from ._discounting_type import *
from ._dividend_extrapolation import *
from ._dividend_type import *
from ._doc_clause import *
from ._double_binary_type import *
from ._end_of_month_convention import *
from ._eti_input_volatility_type import *
from ._exercise_schedule_type import *
from ._exercise_style import *
from ._fixing_frequency import *
from ._format import *
from ._forward_compute_method import *
from ._forward_extrapolation import *
from ._frequency import *
from ._funding_spread_method import *
from ._fx_binary_type import *
from ._fx_cross_type import *
from ._fx_leg_type import *
from ._fx_swap_calculation_method import *
from ._fx_volatility_model import *
from ._holiday_outupts import *
from ._implied_deposit_date_convention import *
from ._in_or_out import *
from ._index_average_method import *
from ._index_compounding_method import *
from ._index_convexity_adjustment_integration_method import *
from ._index_convexity_adjustment_method import *
from ._index_convexity_adjustment_type import *
from ._index_observation_method import *
from ._index_reset_type import *
from ._index_spread_compounding_method import *
from ._inflation_mode import *
from ._input_volatility_type import *
from ._interest_calculation_convention import *
from ._interest_type import *
from ._local_volatility_method import *
from ._method import *
from ._notional_exchange import *
from ._numeraire_type import *
from ._option_volatility_type import *
from ._period_type import *
from ._premium_settlement_type import *
from ._price_side import *
from ._pricing_model_type import *
from ._projected_index_calculation_method import *
from ._quote_fallback_logic import *
from ._redemption_date_type import *
from ._repo_curve_type import *
from ._rounding import *
from ._rounding_type import *
from ._seniority import *
from ._settlement_type import *
from ._status import *
from ._stub_rule import *
from ._swaption_settlement_type import *
from ._swaption_type import *
from ._swaption_volatility_type import *
from ._tenor_reference_date import *
from ._time_stamp import *
from ._underlying_type import *
from ._up_or_down import *
from ._vol_type import *
from ._volatility_adjustment_type import *
from ._volatility_model import *
from ._volatility_term_structure_type import *
from ._volatility_type import *
from ._yield_type import *



import datetime

import pytest

from refinitiv.data._errors import RDError
from refinitiv.data.content.historical_pricing import (
    EventTypes,
    summaries,
    events,
    Intervals,
)
from tests.unit.conftest import StubSession, args
from . import data_for_bypassing_tests as td


@pytest.mark.parametrize(
    "count",
    [
        None,
        6,
        100,
    ],
)
def test_df_intra_bypassing_requests_with_dates_0(count):
    # get all data elements from server

    # given
    session = StubSession(is_open=True, response=td.INTRA_WITH_DATES_0_1)
    expected_str = (
        "AMD.O                OPEN_PRC  HIGH_1  LOW_1  TRDPRC_1  NUM_MOVES  ACVOL_UNS\n"
        "Timestamp                                                                   \n"
        "2022-05-25 23:54:00      90.7   90.77  90.68     90.74         22       4193\n"
        "2022-05-25 23:55:00     90.74   90.74  90.68      90.7         16       1525\n"
        "2022-05-25 23:56:00      90.7    90.8  90.66     90.73         19       2721\n"
        "2022-05-25 23:57:00     90.72   90.75   90.7     90.74         22       3790\n"
        "2022-05-25 23:58:00     90.74   90.75   90.7     90.72         20       1778\n"
        "2022-05-25 23:59:00     90.75   90.75   90.7      90.7         17       2289"
    )
    definition = summaries.Definition(
        universe=["AMD.O"],
        interval=Intervals.MINUTE,
        start="2022-05-25T23:54:00",
        end="2022-05-25T23:59:10",
        count=count,
        fields=["OPEN_PRC", "HIGH_1", "LOW_1", "TRDPRC_1", "NUM_MOVES", "ACVOL_UNS"],
    )
    # when
    response = definition.get_data(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str


def test_df_intra_bypassing_requests_with_dates_1():
    # get only 5 elements from server

    # given
    session = StubSession(is_open=True, response=td.INTRA_WITH_DATES_0_1)
    expected_str = (
        "AMD.O                OPEN_PRC  HIGH_1  LOW_1  TRDPRC_1  NUM_MOVES  ACVOL_UNS\n"
        "Timestamp                                                                   \n"
        "2022-05-25 23:55:00     90.74   90.74  90.68      90.7         16       1525\n"
        "2022-05-25 23:56:00      90.7    90.8  90.66     90.73         19       2721\n"
        "2022-05-25 23:57:00     90.72   90.75   90.7     90.74         22       3790\n"
        "2022-05-25 23:58:00     90.74   90.75   90.7     90.72         20       1778\n"
        "2022-05-25 23:59:00     90.75   90.75   90.7      90.7         17       2289"
    )
    definition = summaries.Definition(
        universe=["AMD.O"],
        interval=Intervals.MINUTE,
        start="2022-05-25T23:54:00",
        end="2022-05-25T23:59:10",
        count=5,
        fields=["OPEN_PRC", "HIGH_1", "LOW_1", "TRDPRC_1", "NUM_MOVES", "ACVOL_UNS"],
    )
    # when
    response = definition.get_data(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str


@pytest.mark.parametrize(
    "finished_datetime",
    [
        datetime.datetime(2022, 5, 25, 23, 56, 59),
        datetime.datetime(2022, 5, 25, 23, 57, 0),
    ],
)
def test_df_bypassing_requests_with_dates_3(finished_datetime):
    # given
    session = StubSession(is_open=True, response=td.INTRA_WITH_DATES_3)
    expected_str = (
        "AMD.O                OPEN_PRC  HIGH_1  LOW_1  TRDPRC_1  NUM_MOVES  ACVOL_UNS\n"
        "Timestamp                                                                   \n"
        "2022-05-25 23:57:00     90.72   90.75   90.7     90.74         22       3790\n"
        "2022-05-25 23:58:00     90.74   90.75   90.7     90.72         20       1778\n"
        "2022-05-25 23:59:00     90.75   90.75   90.7      90.7         17       2289"
    )

    definition = summaries.Definition(
        universe=["AMD.O"],
        interval=Intervals.MINUTE,
        start=finished_datetime,
        end="2022-05-25T23:59:10",
        fields=["OPEN_PRC", "HIGH_1", "LOW_1", "TRDPRC_1", "NUM_MOVES", "ACVOL_UNS"],
    )

    # when
    response = definition.get_data(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str


def test_df_bypassing_requests_with_dates_4():
    # server returns two requests first request > second request

    # given
    session = StubSession(is_open=True, response=td.INTRA_WITH_DATES_4)
    expected_str = (
        "AMD.O                OPEN_PRC  HIGH_1  LOW_1  TRDPRC_1  NUM_MOVES  ACVOL_UNS\n"
        "Timestamp                                                                   \n"
        "2022-05-25 23:54:00      90.7   90.77  90.68     90.74         22       4193\n"
        "2022-05-25 23:55:00     90.74   90.74  90.68      90.7         16       1525\n"
        "2022-05-25 23:56:00      90.7    90.8  90.66     90.73         19       2721\n"
        "2022-05-25 23:57:00     90.72   90.75   90.7     90.74         22       3790"
    )

    definition = summaries.Definition(
        universe=["AMD.O"],
        interval=Intervals.MINUTE,
        start="2022-05-25T23:54:00",
        end="2022-05-25T23:59:10",
        fields=["OPEN_PRC", "HIGH_1", "LOW_1", "TRDPRC_1", "NUM_MOVES", "ACVOL_UNS"],
    )

    # when
    response = definition.get_data(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str


def test_df_intra_bypassing_requests_with_count_0():
    # given
    session = StubSession(is_open=True, response=td.INTRA_WITH_COUNT_0)
    expected_str = (
        "LSEG.L                BID   ASK\n"
        "Timestamp                      \n"
        "2022-10-17 15:34:00  8312  6664\n"
        "2022-10-17 15:35:00  7312  7324\n"
        "2022-10-17 15:40:00  7312  7328\n"
        "2022-10-17 16:30:00  6872  9000\n"
        "2022-10-18 04:00:00  6872  9000"
    )

    definition = summaries.Definition(
        universe=["LSEG.L"],
        interval=Intervals.MINUTE,
        start="10-17-2022T14:30:50",
        count=5,
        fields=["BID", "ASK"],
    )

    # when
    response = definition.get_data(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str


async def test_df_intra_bypassing_requests_async_with_count():
    # given
    session = StubSession(is_open=True, response=td.INTRA_WITH_COUNT_0)
    expected_str = (
        "LSEG.L                BID   ASK\n"
        "Timestamp                      \n"
        "2022-10-17 15:34:00  8312  6664\n"
        "2022-10-17 15:35:00  7312  7324\n"
        "2022-10-17 15:40:00  7312  7328\n"
        "2022-10-17 16:30:00  6872  9000\n"
        "2022-10-18 04:00:00  6872  9000"
    )

    definition = summaries.Definition(
        universe=["LSEG.L"],
        interval=Intervals.MINUTE,
        start="10-17-2022T14:30:50",
        count=5,
        fields=["BID", "ASK"],
    )

    # when
    response = await definition.get_data_async(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str


async def test_df_intra_bypassing_requests_async_with_dates():
    # get only 5 elements from server

    # given
    session = StubSession(is_open=True, response=td.INTRA_WITH_DATES_0_1)
    expected_str = (
        "AMD.O                OPEN_PRC  HIGH_1  LOW_1  TRDPRC_1  NUM_MOVES  ACVOL_UNS\n"
        "Timestamp                                                                   \n"
        "2022-05-25 23:55:00     90.74   90.74  90.68      90.7         16       1525\n"
        "2022-05-25 23:56:00      90.7    90.8  90.66     90.73         19       2721\n"
        "2022-05-25 23:57:00     90.72   90.75   90.7     90.74         22       3790\n"
        "2022-05-25 23:58:00     90.74   90.75   90.7     90.72         20       1778\n"
        "2022-05-25 23:59:00     90.75   90.75   90.7      90.7         17       2289"
    )
    definition = summaries.Definition(
        universe=["AMD.O"],
        interval=Intervals.MINUTE,
        start="2022-05-25T23:54:00",
        end="2022-05-25T23:59:10",
        count=5,
        fields=["OPEN_PRC", "HIGH_1", "LOW_1", "TRDPRC_1", "NUM_MOVES", "ACVOL_UNS"],
    )
    # when
    response = await definition.get_data_async(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str


@pytest.mark.parametrize(
    "object_definition, input_kwargs, response",
    [
        # Summaries
        # one universe
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "start": "9022-05-25T00:00:00",
                "end": "9022-05-25T23:59:10",
                "universe": ["AMD.O"],
                "interval": Intervals.MINUTE,
            },
            response=td.INTRA_EMPTY_DATA,
        ),
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "start": "1022-05-25T00:00:00",
                "end": "1022-05-25T23:59:10",
                "universe": ["AMD.O"],
                "interval": Intervals.MINUTE,
            },
            response=td.INTRA_EMPTY_DATA,
        ),
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "start": "9022-05-25T00:00:00",
                "end": "9022-05-25T23:59:10",
                "count": 1,
                "universe": ["AMD.O"],
                "interval": Intervals.MINUTE,
            },
            response=td.INTRA_EMPTY_DATA,
        ),
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "start": "1022-05-25T00:00:00",
                "end": "1022-05-25T23:59:10",
                "count": 1,
                "universe": ["AMD.O"],
                "interval": Intervals.MINUTE,
            },
            response=td.INTRA_EMPTY_DATA,
        ),
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "end": "1022-05-25T23:59:10",
                "count": 1,
                "universe": ["AMD.O"],
                "interval": Intervals.MINUTE,
            },
            response=td.INTRA_EMPTY_DATA,
        ),
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "end": "1022-05-25T23:59:10",
                "count": 1,
                "universe": ["AMD.O"],
                "interval": Intervals.MINUTE,
            },
            response=td.INTRA_EMPTY_DATA,
        ),
        # two universes
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "start": "9022-05-25T00:00:00",
                "end": "9022-05-25T23:59:10",
                "universe": ["AMD.O", "VOD.L"],
                "interval": Intervals.MINUTE,
            },
            response=td.INTRA_TWO_UNIVERSES_EMPTY_DATA,
        ),
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "start": "1022-05-25T00:00:00",
                "end": "1022-05-25T23:59:10",
                "universe": ["AMD.O", "VOD.L"],
                "interval": Intervals.MINUTE,
            },
            response=td.INTRA_TWO_UNIVERSES_EMPTY_DATA,
        ),
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "start": "9022-05-25T00:00:00",
                "end": "9022-05-25T23:59:10",
                "count": 1,
                "universe": ["AMD.O", "VOD.L"],
                "interval": Intervals.MINUTE,
            },
            response=td.INTRA_TWO_UNIVERSES_EMPTY_DATA,
        ),
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "start": "1022-05-25T00:00:00",
                "end": "1022-05-25T23:59:10",
                "count": 1,
                "universe": ["AMD.O", "VOD.L"],
                "interval": Intervals.MINUTE,
            },
            response=td.INTRA_TWO_UNIVERSES_EMPTY_DATA,
        ),
        # Events
        # one universe
        args(
            object_definition=events.Definition,
            input_kwargs={
                "start": "9022-05-25T00:00:00",
                "end": "9022-05-25T23:59:10",
                "universe": ["AMD.O"],
            },
            response=td.INTRA_EMPTY_DATA,
        ),
        args(
            object_definition=events.Definition,
            input_kwargs={
                "start": "1022-05-25T00:00:00",
                "end": "1022-05-25T23:59:10",
                "universe": ["AMD.O"],
            },
            response=td.INTRA_EMPTY_DATA,
        ),
        args(
            object_definition=events.Definition,
            input_kwargs={
                "start": "9022-05-25T00:00:00",
                "end": "9022-05-25T23:59:10",
                "count": 10001,
                "universe": ["AMD.O"],
            },
            response=td.INTRA_EMPTY_DATA,
        ),
        args(
            object_definition=events.Definition,
            input_kwargs={
                "start": "1022-05-25T00:00:00",
                "end": "1022-05-25T23:59:10",
                "count": 10001,
                "universe": ["AMD.O"],
            },
            response=td.INTRA_EMPTY_DATA,
        ),
        args(
            object_definition=events.Definition,
            input_kwargs={
                "end": "1022-05-25T23:59:10",
                "count": 10001,
                "universe": ["AMD.O"],
            },
            response=td.INTRA_EMPTY_DATA,
        ),
        args(
            object_definition=events.Definition,
            input_kwargs={
                "end": "1022-05-25T23:59:10",
                "count": 10001,
                "universe": ["AMD.O"],
            },
            response=td.INTRA_EMPTY_DATA,
        ),
        # two universes
        args(
            object_definition=events.Definition,
            input_kwargs={
                "start": "9022-05-25T00:00:00",
                "end": "9022-05-25T23:59:10",
                "universe": ["AMD.O", "VOD.L"],
            },
            response=td.INTRA_TWO_UNIVERSES_EMPTY_DATA,
        ),
        args(
            object_definition=events.Definition,
            input_kwargs={
                "start": "1022-05-25T00:00:00",
                "end": "1022-05-25T23:59:10",
                "universe": ["AMD.O", "VOD.L"],
            },
            response=td.INTRA_TWO_UNIVERSES_EMPTY_DATA,
        ),
        args(
            object_definition=events.Definition,
            input_kwargs={
                "start": "9022-05-25T00:00:00",
                "end": "9022-05-25T23:59:10",
                "count": 10001,
                "universe": ["AMD.O", "VOD.L"],
            },
            response=td.INTRA_TWO_UNIVERSES_EMPTY_DATA,
        ),
        args(
            object_definition=events.Definition,
            input_kwargs={
                "start": "1022-05-25T00:00:00",
                "end": "1022-05-25T23:59:10",
                "count": 10001,
                "universe": ["AMD.O", "VOD.L"],
            },
            response=td.INTRA_TWO_UNIVERSES_EMPTY_DATA,
        ),
    ],
)
def test_df_intra_server_returns_empty_data(object_definition, input_kwargs, response):
    # given
    session = StubSession(is_open=True, response=response)
    expected_str = "Empty DataFrame\n" "Columns: []\n" "Index: []"

    definition = object_definition(**input_kwargs)
    # when
    response = definition.get_data(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str


@pytest.mark.parametrize(
    "object_definition, input_kwargs, response",
    [
        # Summaries
        # one universe
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "start": "2022-05-25T23:54:00",
                "end": "2022-05-25T23:59:10",
                "universe": ["LSEG.L.ffff"],
                "interval": Intervals.MINUTE,
            },
            response=td.INST_IS_BAD,
        ),
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "start": "2022-05-25T23:54:00",
                "count": 1,
                "universe": ["LSEG.L.ffff"],
                "interval": Intervals.MINUTE,
            },
            response=td.INST_IS_BAD,
        ),
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "end": "2022-05-25T23:59:10",
                "count": 1,
                "universe": ["LSEG.L.ffff"],
                "interval": Intervals.MINUTE,
            },
            response=td.INST_IS_BAD,
        ),
        # two universes
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "start": "2022-05-25T23:54:00",
                "end": "2022-05-25T23:59:10",
                "universe": ["LSEG.L.ffff", "VOD.Lffff"],
                "interval": Intervals.MINUTE,
            },
            response=td.INSTS_ARE_BAD,
        ),
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "start": "2022-05-25T23:54:00",
                "count": 1,
                "universe": ["LSEG.L.ffff", "VOD.Lffff"],
                "interval": Intervals.MINUTE,
            },
            response=td.INSTS_ARE_BAD,
        ),
        args(
            object_definition=summaries.Definition,
            input_kwargs={
                "end": "2022-05-25T23:59:10",
                "count": 1,
                "universe": ["LSEG.L.ffff", "VOD.Lffff"],
                "interval": Intervals.MINUTE,
            },
            response=td.INSTS_ARE_BAD,
        ),
        # Events
        # one universe
        args(
            object_definition=events.Definition,
            input_kwargs={
                "start": "2022-05-25T23:54:00",
                "end": "2022-05-25T23:59:10",
                "universe": ["LSEG.L.ffff"],
            },
            response=td.INST_IS_BAD,
        ),
        args(
            object_definition=events.Definition,
            input_kwargs={
                "start": "2022-05-25T23:54:00",
                "count": 10001,
                "universe": ["LSEG.L.ffff"],
            },
            response=td.INST_IS_BAD,
        ),
        args(
            object_definition=events.Definition,
            input_kwargs={
                "end": "2022-05-25T23:59:10",
                "count": 10001,
                "universe": ["LSEG.L.ffff"],
            },
            response=td.INST_IS_BAD,
        ),
        # two universes
        args(
            object_definition=events.Definition,
            input_kwargs={
                "start": "2022-05-25T23:54:00",
                "end": "2022-05-25T23:59:10",
                "universe": ["LSEG.L.ffff", "VOD.Lffff"],
            },
            response=td.INSTS_ARE_BAD,
        ),
        args(
            object_definition=events.Definition,
            input_kwargs={
                "start": "2022-05-25T23:54:00",
                "count": 10001,
                "universe": ["LSEG.L.ffff", "VOD.Lffff"],
            },
            response=td.INSTS_ARE_BAD,
        ),
        args(
            object_definition=events.Definition,
            input_kwargs={
                "end": "2022-05-25T23:59:10",
                "count": 10001,
                "universe": ["LSEG.L.ffff", "VOD.Lffff"],
            },
            response=td.INSTS_ARE_BAD,
        ),
    ],
)
def test_df_intra_server_returns_error_data(object_definition, input_kwargs, response):
    # given
    session = StubSession(is_open=True, response=response)
    error_msg = "Error code 1 | No data to return, please check errors: ERROR: No successful response."

    definition = object_definition(**input_kwargs)

    # when
    with pytest.raises(RDError, match=error_msg):
        definition.get_data(session=session)


@pytest.mark.parametrize(
    "count",
    [
        None,
        6,
        100,
    ],
)
def test_df_events_bypassing_requests_with_dates_0(count):
    # get all data elements from server

    # given
    session = StubSession(is_open=True, response=td.EVENT_WITH_DATES_0_1)
    expected_str = (
        "GBP=                    EVENT_TYPE    RTL     BID     ASK  MID_PRICE  DSPLY_NAME SRC_REF1 DLG_CODE1       CTBTR_1 CTB_LOC1  QUALIFIERS\n"
        "Timestamp                                                                                                                             \n"
        "2022-10-18 09:17:03.269      quote  25886  1.1318  1.1321    1.13195        <NA>     COB1      CBKF   Commerzbank      FFT        <NA>\n"
        "2022-10-18 09:17:04.118      quote  25950  1.1318  1.1322      1.132        <NA>     BCFX      <NA>      BARCLAYS      LON        <NA>\n"
        "2022-10-18 09:17:04.310      quote  26014  1.1318   1.132     1.1319        <NA>     <NA>      ZKBZ   ZUERCHER KB      ZUR        <NA>\n"
        "2022-10-18 09:17:04.416      quote  26078  1.1315   1.132    1.13175        <NA>     CKLU      CKLU  StoneX Finan      LUX        <NA>\n"
        "2022-10-18 09:17:04.466      quote  26142  1.1319   1.132    1.13195        <NA>     SAHK      SAHK     SANTANDER      HKG        <NA>"
    )

    definition = events.Definition(
        universe=["GBP="],
        eventTypes=[EventTypes.TRADE, EventTypes.QUOTE],
        start="10-17-2022T00:00:00",
        end="10-18-2022T10:00:00",
        count=count,
    )

    # when
    response = definition.get_data(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str


def test_df_events_bypassing_requests_with_dates_1():
    # get only 3 elements

    # given
    session = StubSession(is_open=True, response=td.EVENT_WITH_DATES_0_1)
    expected_str = (
        "GBP=                    EVENT_TYPE    RTL     BID    ASK  MID_PRICE  DSPLY_NAME SRC_REF1 DLG_CODE1       CTBTR_1 CTB_LOC1  QUALIFIERS\n"
        "Timestamp                                                                                                                            \n"
        "2022-10-18 09:17:04.310      quote  26014  1.1318  1.132     1.1319        <NA>     <NA>      ZKBZ   ZUERCHER KB      ZUR        <NA>\n"
        "2022-10-18 09:17:04.416      quote  26078  1.1315  1.132    1.13175        <NA>     CKLU      CKLU  StoneX Finan      LUX        <NA>\n"
        "2022-10-18 09:17:04.466      quote  26142  1.1319  1.132    1.13195        <NA>     SAHK      SAHK     SANTANDER      HKG        <NA>"
    )

    definition = events.Definition(
        universe=["GBP="],
        eventTypes=[EventTypes.TRADE, EventTypes.QUOTE],
        start="10-17-2022T00:00:00",
        end="10-18-2022T10:00:00",
        count=3,
    )

    # when
    response = definition.get_data(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str


@pytest.mark.asyncio
async def test_df_events_bypassing_requests_async_with_dates():
    # get only 3 elements

    # given
    session = StubSession(is_open=True, response=td.EVENT_WITH_DATES_0_1)
    expected_str = (
        "GBP=                    EVENT_TYPE    RTL     BID    ASK  MID_PRICE  DSPLY_NAME SRC_REF1 DLG_CODE1       CTBTR_1 CTB_LOC1  QUALIFIERS\n"
        "Timestamp                                                                                                                            \n"
        "2022-10-18 09:17:04.310      quote  26014  1.1318  1.132     1.1319        <NA>     <NA>      ZKBZ   ZUERCHER KB      ZUR        <NA>\n"
        "2022-10-18 09:17:04.416      quote  26078  1.1315  1.132    1.13175        <NA>     CKLU      CKLU  StoneX Finan      LUX        <NA>\n"
        "2022-10-18 09:17:04.466      quote  26142  1.1319  1.132    1.13195        <NA>     SAHK      SAHK     SANTANDER      HKG        <NA>"
    )

    definition = events.Definition(
        universe=["GBP="],
        eventTypes=[EventTypes.TRADE, EventTypes.QUOTE],
        start="10-17-2022T00:00:00",
        end="10-18-2022T10:00:00",
        count=3,
    )

    # when
    response = await definition.get_data_async(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str



from refinitiv.data.content.ipa._curves._cross_currency_curves._curves import (
    FxForwardCurveDefinition,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._curves import (
    CrossCurrencyCurveDefinitionPricing,
)
from refinitiv.data.content.ipa._curves import ZcCurveDefinition


def test_fx_forward_curve_definition():
    # given
    expected_dict = {
        "baseCurrency": "string",
        "baseIndexName": "string",
        "crossCurrencyDefinitions": [{}],
        "curveTenors": ["string1", "string2"],
        "interestRateCurveDefinitions": [{}],
        "isNonDeliverable": True,
        "pivotCurrency": "string",
        "pivotIndexName": "string",
        "quotedCurrency": "string",
        "quotedIndexName": "string",
    }

    # when
    testing_obj = FxForwardCurveDefinition(
        cross_currency_definitions=[CrossCurrencyCurveDefinitionPricing()],
        curve_tenors=["string1", "string2"],
        interest_rate_curve_definitions=[ZcCurveDefinition()],
        base_currency="string",
        base_index_name="string",
        is_non_deliverable=True,
        pivot_currency="string",
        pivot_index_name="string",
        quoted_currency="string",
        quoted_index_name="string",
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.cross_currency_definitions == [
        CrossCurrencyCurveDefinitionPricing()
    ]
    assert testing_obj.curve_tenors == ["string1", "string2"]
    assert testing_obj.interest_rate_curve_definitions == [ZcCurveDefinition()]
    assert testing_obj.base_currency == "string"
    assert testing_obj.base_index_name == "string"
    assert testing_obj.is_non_deliverable is True
    assert testing_obj.pivot_currency == "string"
    assert testing_obj.pivot_index_name == "string"
    assert testing_obj.quoted_currency == "string"
    assert testing_obj.quoted_index_name == "string"



# coding: utf8

__all__ = ["StrikeFilterRange"]

from ..._object_definition import ObjectDefinition


class StrikeFilterRange(ObjectDefinition):
    def __init__(
        self,
        max_of_median_implied_vol_percent=None,
        min_of_median_implied_vol_percent=None,
    ):
        super().__init__()
        self.max_of_median_implied_vol_percent = max_of_median_implied_vol_percent
        self.min_of_median_implied_vol_percent = min_of_median_implied_vol_percent

    @property
    def max_of_median_implied_vol_percent(self):
        """
        Remove strikes whose implied vol is more than MaxOfMedianImpliedVolPercent x Median implied Vol.
        :return: float
        """
        return self._get_parameter("maxOfMedianImpliedVolPercent")

    @max_of_median_implied_vol_percent.setter
    def max_of_median_implied_vol_percent(self, value):
        self._set_parameter("maxOfMedianImpliedVolPercent", value)

    @property
    def min_of_median_implied_vol_percent(self):
        """
        Remove strikes whose implied vol is less than MinOfMedianImpliedVolPercent x Median implied Vol.
        :return: float
        """
        return self._get_parameter("minOfMedianImpliedVolPercent")

    @min_of_median_implied_vol_percent.setter
    def min_of_median_implied_vol_percent(self, value):
        self._set_parameter("minOfMedianImpliedVolPercent", value)



from dataclasses import dataclass


@dataclass
class Credentials:
    username: str
    password: str


config_for_beta_user = {
    "app_key": "PPE_DESKTOP_APP_KEY",
    "username": "PPE_USERNAME",
    "password": "PPE_PASSWORD",
}

config_for_prod_user = {
    "app_key": "DESKTOP_APP_KEY",
    "username": "EDP_USERNAME",
    "password": "EDP_PASSWORD",
    "username_second": "EDP_USERNAME_2",
    "password_second": "EDP_PASSWORD_2",
    "username_third": "EDP_USERNAME_3",
    "password_third": "EDP_PASSWORD_3",
    "rdp_username": "RDP_USERNAME",
    "rdp_password": "RDP_PASSWORD",
    "rdp_username_second": "RDP_USERNAME_2",
    "rdp_password_second": "RDP_PASSWORD_2",
    "eikon_username": "EIKON_USERNAME",
    "eikon_password": "EIKON_PASSWORD",
}



from refinitiv.data.content.ipa.surfaces import fx


def surface_fx_definition():
    definition = fx.Definition(
        underlying_definition={"fxCrossCode": "EURUSD"},
        surface_tag="FxVol-EURUSD",
        surface_layout=fx.SurfaceLayout(format=fx.Format.MATRIX),
        surface_parameters=fx.FxCalculationParams(
            x_axis=fx.Axis.DATE,
            y_axis=fx.Axis.STRIKE,
            calculation_date="2018-08-20T00:00:00Z",
        ),
    )
    return definition


def invalid_surface_fx_definition():
    definition = fx.Definition(
        underlying_definition={"fxCrossCode": "INVAL"},
        surface_tag="FxVol-EURUSD",
        surface_layout=fx.SurfaceLayout(format=fx.Format.MATRIX),
        surface_parameters=fx.FxCalculationParams(
            x_axis=fx.Axis.DATE,
            y_axis=fx.Axis.STRIKE,
            calculation_date="2018-08-20T00:00:00Z",
        ),
    )
    return definition



import allure
import pytest

from refinitiv.data.content.ownership.fund import holdings
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ownership.helpers import (
    check_response_dataframe_instrument_contains_expected_universes,
    get_expected_column,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_extended_params_were_sent_in_request,
    check_dataframe_column_date_for_datetime_type,
)

EXPECTED_COLUMN_NAMES = [
    "instrument",
    "TR.FundName",
    "Total Holdings Count",
    "TR.FdSecurityOwnedRIC",
    "TR.FdSecurityOwnedName",
    "TR.FdInvPortfolioRank",
    "TR.FdInvestorSharesHeld",
    "TR.FdPctSecuritySharesOut",
    "TR.FdInvestorPctPortfolio",
    "TR.FdInvestorValueHeld",
    "TR.FdInvestorFilingDate",
    "TR.FdInvestorFilingType",
    "TR.FdInvestorShrsHeldChg",
    "TR.FdInvestorValueHeldChg",
    "TR.FdInvestorPosPctChg",
]

EXPECTED_COLUMN_TITLES = [
    "Instrument",
    "Name",
    "Total Holdings Count",
    "RIC",
    "Security Owned Name",
    "Portfolio Rank",
    "SharesHeld",
    "% SharesOutstanding",
    "% Portfolio",
    "SharesHeld Value",
    "Holdings Date",
    "Filing Type",
    "SharesHeld Change",
    "SharesHeld Value Change",
    "SharesHeld Change %",
]


@allure.suite("Content object - Ownership Fund Holdings")
@allure.feature("Content object - Ownership Fund Holdings")
@allure.severity(allure.severity_level.CRITICAL)
class TestFundHoldings:
    @allure.title(
        "Create Fund Holdings definition object with valid params and get data"
    )
    @pytest.mark.caseid("37478861")
    @pytest.mark.parametrize(
        "universe,limit,use_field_names_in_headers,extended_params,expected_universe",
        [
            ("ONEX.CCP", 5, True, {"test_param": "test_value"}, "ONEX.CCP"),
            (
                ["TRI.N", "ONEX.CCP", "AAPL.O"],
                100,
                False,
                {"universe": "TRI.N,US02079K1079"},
                ["TRI.N", "US02079K1079"],
            ),
        ],
    )
    def test_create_fund_holdings_object_with_valid_fields_and_get_data(
        self,
        open_platform_session,
        universe,
        limit,
        use_field_names_in_headers,
        extended_params,
        expected_universe,
    ):
        expected_column = get_expected_column(
            use_field_names_in_headers, EXPECTED_COLUMN_TITLES, EXPECTED_COLUMN_NAMES
        )
        response = holdings.Definition(
            universe, limit, use_field_names_in_headers, extended_params
        ).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(response, expected_column)
        check_response_dataframe_instrument_contains_expected_universes(
            response, expected_universe
        )
        check_dataframe_column_date_for_datetime_type(response)
        check_extended_params_were_sent_in_request(response, extended_params)

    @allure.title(
        "Create Fund Holdings definition object with mix valid and invalid params and get data asynchronously"
    )
    @pytest.mark.caseid("37478862")
    @pytest.mark.parametrize(
        "universe,invalid_universe",
        [
            (["LP40189339", "invalid universe"], "invalid universe"),
        ],
    )
    @pytest.mark.smoke
    async def test_create_fund_holdings_object_with_mix_valid_and_invalid_params_and_get_data_async(
        self, open_platform_session_async, universe, invalid_universe
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            holdings.Definition(universe, use_field_names_in_headers=True),
            holdings.Definition(invalid_universe, use_field_names_in_headers=True),
        )
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code=None,
            expected_error_message="Failed to resolve identifiers ['invalid universe']",
        )
        check_non_empty_response_data(valid_response)
        check_response_dataframe_contains_columns_names(
            valid_response, EXPECTED_COLUMN_NAMES
        )
        check_response_dataframe_instrument_contains_expected_universes(
            valid_response, universe
        )



from typing import Any, TYPE_CHECKING

import pandas as pd

from ...._content_type import ContentType
from ...._tools import cached_property
from ....delivery._stream import _RDPStream, StreamStateEvent, StreamStateManager
from ....delivery._stream._stream_factory import create_rdp_stream
from ....delivery._stream._stream_listener import RDPStreamListener

if TYPE_CHECKING:
    from ...._types import ExtendedParams
    from ...._core.session import Session
    from ....delivery._stream import StreamState


class QuantitativeDataStream(
    StreamStateManager, RDPStreamListener["QuantitativeDataStream"]
):
    def __init__(
        self,
        universe: dict,
        session: "Session",
        fields: list = None,
        extended_params: "ExtendedParams" = None,
    ):
        self._session = session

        StreamStateManager.__init__(self, logger=self._session.logger())
        RDPStreamListener.__init__(self, logger=self._session.logger())

        self._universe = universe
        self._fields = fields
        self._extended_params = extended_params

        if extended_params and "view" in extended_params:
            self._column_names = extended_params["view"]

        else:
            self._column_names = fields or None

        self._data = None
        self._headers = None

    @property
    def universe(self):
        return self._universe

    @property
    def session(self) -> "Session":
        return self._session

    @session.setter
    def session(self, session: "Session"):
        if self._session != session and not self.is_open:
            self._session = session
            RDPStreamListener.init_logger(self, self._session.logger())
            self._stream.session = self._session

    @cached_property
    def _stream(self) -> _RDPStream:
        stream = create_rdp_stream(
            ContentType.STREAMING_CONTRACTS,
            session=self._session,
            view=self._fields,
            universe=self._universe,
            extended_params=self._extended_params,
        )
        stream.on(StreamStateEvent.CLOSED, lambda *_: self.close())
        stream.on_ack(self._on_stream_ack)
        stream.on_response(self._on_stream_response)
        stream.on_update(self._on_stream_update)
        stream.on_alarm(self._on_stream_alarm)
        return stream

    def __repr__(self):
        s = super().__repr__()
        s = s.replace(">", f" {{name='{self._universe}', state={self._stream.state}}}>")
        return s

    @property
    def df(self):
        if self._data is None or self._column_names is None:
            return pd.DataFrame([])
        return pd.DataFrame.from_records(self._data, columns=self._column_names)

    def get_snapshot(self) -> pd.DataFrame:
        return self.df

    def _do_open(self, **kwargs) -> "StreamState":
        return self._stream.open(**kwargs)

    def _do_close(self, **kwargs) -> "StreamState":
        return self._stream.close(**kwargs)

    def _do_on_stream_ack(self, stream: _RDPStream, message: dict) -> Any:
        return message["state"]

    def _do_on_stream_response(self, stream: _RDPStream, message: dict) -> Any:
        if "data" in message:
            self._data = message["data"]

        if "headers" in message:
            self._headers = message["headers"]
            self._column_names = [col["name"] for col in self._headers]

        return self._data, self._column_names

    def _do_on_stream_update(self, stream: _RDPStream, message: dict) -> Any:
        if "data" in message:
            self._data = message["data"]

        return self._data, self._column_names

    def _do_on_stream_alarm(self, stream: _RDPStream, message: dict) -> Any:
        return message["state"]



import threading
from collections import defaultdict
from types import SimpleNamespace
from typing import Union, List, Optional

import pytest

import refinitiv.data as rd
from refinitiv.data._content_type import ContentType
from refinitiv.data.delivery._data._api_type import APIType
from refinitiv.data.delivery._stream import (
    StreamConnection,
    RDPStreamConnection,
    OMMStreamConnection,
)
from refinitiv.data.delivery._stream._protocol_type import ProtocolType
from refinitiv.data.delivery._stream._stream_cxn_config_data import (
    PlatformStreamCxnConfig,
    StreamServiceInfo,
)
from refinitiv.data.delivery._stream._stream_factory import StreamDetails
from refinitiv.data.delivery._stream.contrib import ContribResponse
from refinitiv.data.delivery._stream.contrib._offstream import _OffStreamContrib
from refinitiv.data.delivery._stream.contrib._response import (
    AckContribResponse,
    NullContribResponse,
)
from tests.unit.conftest import StubSession


@pytest.fixture(
    scope="function",
    params=[
        rd.content.pricing.Definition("universe"),
        rd.content.trade_data_service.Definition(),
        rd.delivery.rdp_stream.Definition("", [], [], {}, "api/path"),
        rd.delivery.omm_stream.Definition(""),
        rd.content.pricing.chain.Definition("universe"),
        rd.content.ipa.financial_contracts.bond.Definition("universe"),
        rd.content.custom_instruments.Definition("universe"),
    ],
    ids=[
        "pricing",
        "trade_data_service",
        "rdp_stream",
        "omm_stream",
        "pricing.chain",
        "financial_contracts.bond",
        "custom_instruments",
    ],
)
def stream(request):
    definition = request.param
    session = StubSession(is_open=True)
    stream = definition.get_stream(session)
    return stream


class StubWebSocketApp:
    def __init__(self, **kwargs) -> None:
        self.url = kwargs["url"]
        self.on_open = kwargs["on_open"]
        self.on_message = kwargs["on_message"]
        self.on_error = kwargs["on_error"]
        self.on_close = kwargs["on_close"]
        self.on_ping = kwargs["on_ping"]
        self.on_pong = kwargs["on_pong"]

        self._is_close = False
        self._cmdname = ""
        config_by_cmdname = {
            "on_open": {"func": self.on_open, "args": []},
            "on_message": {"func": self.on_message, "args": []},
            "on_error": {"func": self.on_error, "args": [Exception()]},
            "on_close": {"func": self._do_on_close, "args": ["", ""]},
            "on_ping": {"func": self.on_ping, "args": [{}]},
            "on_pong": {"func": self.on_pong, "args": [{}]},
        }
        self._config_by_cmdname = config_by_cmdname
        self._func_called = threading.Event()
        self._cmd = threading.Event()
        self._args_for_cmd = None

    def cmd(self, name, *args):
        cmds = self._config_by_cmdname.keys()
        if name in cmds:
            self._func_called.clear()
            self._args_for_cmd = args
            self._cmdname = name
            self._cmd.set()
            self._func_called.wait()
        else:
            raise ValueError(f"No such command: {name}, has: {cmds}")

    def _cmd_applet(self):
        while not self._is_close:
            self._cmd.wait()
            if self._is_close:
                break
            config = self._config_by_cmdname[self._cmdname]
            func = config["func"]
            args = self._args_for_cmd or config["args"]
            try:
                func(self, *args)
            except Exception as e:
                print(e)
            self._func_called.set()
            self._cmdname = ""
            if not self._is_close:
                self._cmd.clear()

    def run_forever(self, *args, **kwargs):
        t = threading.Thread(target=self._cmd_applet)
        t.start()
        t.join()

    def send(self, s):
        # do nothing
        pass

    def close(self):
        self._is_close = True
        self._cmd.set()
        self._func_called.set()

    def _do_on_close(self, ws, close_status_code: str, close_msg: str):
        self.on_close(ws, close_status_code, close_msg)
        self.close()


class StubStreamCxn(StreamConnection):
    def send(self, data: dict) -> None:
        # do nothing
        pass


class StubStreamCxnConfig(PlatformStreamCxnConfig):
    def __init__(
        self,
        infos: Union[List["StreamServiceInfo"], "StreamServiceInfo"] = None,
        protocols: Union[List[str], str] = None,
        transport: str = "websocket",
    ):
        infos = infos or [StreamServiceInfo("", "", "", "", "", "", "")]
        protocols = protocols or [""]
        super().__init__(infos, protocols, transport)


def create_stream_cxn(session, details):
    return StubStreamCxn(1, "name", session, {}, 1)


class StubStreamConnection:
    def __init__(self, *args, **kwargs) -> None:
        self.id = "id"
        self.name = "name"
        self.daemon = True
        self._connection_result_ready = threading.Event()
        self.is_disposed = False
        self.is_disconnecting = False
        self.is_disconnected = False
        self._callbacks = defaultdict(list)
        self.session = StubSession()

    def wait_connection_result(self):
        # do nothing
        pass

    def start(self):
        self._connection_result_ready.set()

    def on(self, event, callback):
        self._callbacks[event].append(callback)

    def send_message(self, msg):
        for event, callbacks in self._callbacks.items():
            if (
                "complete_stream" in event
                or "response_stream" in event
                or "refresh_stream" in event
            ):
                for cb in callbacks.copy():
                    try:
                        cb(self, {"key": "value"})
                    except:
                        pass

    def remove_listener(self, *args, **kwargs):
        pass

    def dispose(self, *args, **kwargs):
        pass

    def join(self, *args, **kwargs):
        pass


def dummy(*args, **kwargs):
    pass


timer_mock = SimpleNamespace()
timer_mock.wait = dummy

event_mock = SimpleNamespace()
event_mock.wait = dummy
event_mock.clear = dummy


@pytest.fixture(scope="function")
def create_rdp_cxn():
    cxn: Optional[RDPStreamConnection] = None

    def creator(stream_auto_reconnection=False) -> RDPStreamConnection:
        nonlocal cxn
        sess = StubSession(stream_auto_reconnection=stream_auto_reconnection)
        cfg = StubStreamCxnConfig()
        cxn = RDPStreamConnection(
            connection_id=1, name="fixture-name", session=sess, config=cfg
        )
        return cxn

    yield creator

    if not cxn.is_disposed:
        cxn.start_disconnecting()
        cxn.end_disconnecting()
        cxn.dispose()
        cxn.join()

    assert cxn.is_disposed, cxn.state


@pytest.fixture(scope="function")
def create_omm_cxn():
    cxn: Optional[OMMStreamConnection] = None

    def creator(stream_auto_reconnection=False) -> OMMStreamConnection:
        nonlocal cxn
        sess = StubSession(stream_auto_reconnection=stream_auto_reconnection)
        cfg = StubStreamCxnConfig()
        cxn = OMMStreamConnection(
            connection_id=1, name="fixture-name", session=sess, config=cfg
        )
        return cxn

    yield creator

    if not cxn.is_disposed:
        cxn.start_disconnecting()
        cxn.end_disconnecting()
        cxn.dispose()
        cxn.join()

    assert cxn.is_disposed, cxn.state


class StubOffStreamContrib(_OffStreamContrib):
    def __init__(self, session, raise_err=False, was_send=True) -> None:
        self.was_send = was_send
        self.raise_err = raise_err

        details = StreamDetails(
            ContentType.STREAMING_OFF_CONTRIB,
            ProtocolType.OMM_OFF_CONTRIB,
            APIType.STREAMING_CONTRIB,
        )
        super().__init__(1, session, "name", details, "service", "domain")

    def _do_open(self, *args, **kwargs):
        if self.raise_err:
            raise ConnectionError("Error message here")

    def send(self, data: dict) -> bool:
        return self.was_send

    def _dispose(self):
        pass

    def contribute(
        self,
        fields: dict,
        contrib_type: Union[str, "ContribType", None] = None,
        post_user_info: Optional[dict] = None,
    ) -> ContribResponse:
        if self.was_send:
            return AckContribResponse(
                {
                    "ID": 2,
                    "Type": "Ack",
                    "AckID": 5,
                    "Text": "[1]: Contribution Accepted",
                    "Key": {"Service": "ATS_GLOBAL_1", "Name": "TEST"},
                }
            )
        else:
            return NullContribResponse()

    async def contribute_async(
        self,
        fields: dict,
        contrib_type: Union[str, "ContribType", None] = None,
        post_user_info: Optional[dict] = None,
    ) -> ContribResponse:
        if self.was_send:
            return AckContribResponse(
                {
                    "ID": 2,
                    "Type": "Ack",
                    "AckID": 5,
                    "Text": "[1]: Contribution Accepted",
                    "Key": {"Service": "ATS_GLOBAL_1", "Name": "TEST"},
                }
            )
        else:
            return NullContribResponse()


def mock_create_offstream_contrib(session, *args, **kwargs):
    return StubOffStreamContrib(session)


def make_mock_create_offstream_contrib(raise_err=False, was_send=True):
    def inner(session, *args, **kwargs):
        return StubOffStreamContrib(session, raise_err, was_send)

    return inner



from datetime import timedelta, datetime

import allure
import pytest

from refinitiv.data.content import custom_instruments as ci
from refinitiv.data.content.custom_instruments._custom_instruments_data_provider import (
    get_user_id,
)
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.helpers import (
    get_async_response_from_definitions,
    assert_error,
    check_response_status,
    check_non_empty_response_data,
    check_response_data_start_end_date,
    is_universe_empty,
    check_response_is_success,
    check_response_data_universe,
    is_expected_param_in_request_url,
    check_index_column_contains_dates,
)


@allure.suite("Content object - Custom Instrument")
@allure.feature("Custom Instrument Events")
@allure.severity(allure.severity_level.CRITICAL)
class TestCustomInstrumentEvents:
    @allure.title("Create Custom Instrument events definition and get data")
    @pytest.mark.caseid(["C42988157", "C42988703"])
    @pytest.mark.smoke
    def test_custom_instrument_events_and_get_data(
        self, open_session_with_rdp_creds_for_ci, create_instrument
    ):
        instrument = create_instrument()
        response = ci.events.Definition(universe=instrument, count=15).get_data()
        check_response_is_success(response)
        check_non_empty_response_data(response)
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        assert len(response.data.raw["data"]) == 15, (
            f"Custom Instrument event dataframe size is not as expected: "
            f"{response.data.raw['data']} != {15}"
        )
        check_index_column_contains_dates(response)

    @allure.title(
        "Create Custom Instrument multi-request events definition and get data with interval"
    )
    @pytest.mark.caseid("C42991351")
    @pytest.mark.parametrize(
        "start_date, end_date",
        [
            (timedelta(hours=-10), timedelta(0)),
            (timedelta(hours=3), datetime.now()),
            ("2022-12-08T05:00:00", "2022-12-09T07:10:00"),
        ],
        ids=["timedelta", "datetime", "string"],
    )
    def test_custom_instrument_events_get_data_with_interval(
        self, open_desktop_session, create_instrument, start_date, end_date, request
    ):
        if "datetime" in request.node.callspec.id:
            start_date = end_date - start_date

        instrument_01 = create_instrument()
        instrument_02 = create_instrument()
        response = ci.events.Definition(
            universe=[instrument_01, instrument_02],
            start=start_date,
            end=end_date,
            count=13000,
        ).get_data()
        is_expected_param_in_request_url(response, ["start", "end"])
        check_response_is_success(response)
        check_non_empty_response_data(response)
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_response_data_universe(response, [instrument_01, instrument_02])
        check_response_data_start_end_date(response, start_date, end_date, request)
        check_index_column_contains_dates(response)

    @allure.title(
        "Create Custom Instrument events definition get data with mix of valid and invalid universe"
    )
    @pytest.mark.caseid("C42991486")
    def test_custom_instrument_events_get_data_with_mix_of_valid_and_invalid_universe(
        self, open_desktop_session, create_instrument
    ):
        instrument = create_instrument()
        response = ci.events.Definition([instrument, "invalid_universe"]).get_data()
        check_response_is_success(response)
        assert not is_universe_empty(response.data.df, instrument)
        assert is_universe_empty(response.data.df, "invalid_universe")

    @allure.title("Custom Instrument events definition - get data without universe")
    @pytest.mark.caseid("C42991497")
    def test_custom_instrument_events_get_data_without_universe(self):
        with pytest.raises(TypeError) as error:
            ci.events.Definition()
        assert_error(error, "universe")

    @allure.title(
        "Create Custom Instrument events get data async with valid and invalid universe"
    )
    @pytest.mark.caseid("C42991499")
    async def test_custom_instrument_events_get_data_async(
        self, open_desktop_session_async, create_instrument
    ):
        instrument = create_instrument()
        uuid = get_user_id()
        valid_response, invalid_response = await get_async_response_from_definitions(
            ci.events.Definition(instrument),
            ci.events.Definition("invalid_universe"),
        )
        check_response_is_success(valid_response)
        check_non_empty_response_data(valid_response)
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.FOUR_HUNDRED_FOUR,
            expected_http_reason=HttpReason.NOT_FOUND,
            expected_error_code="404",
            expected_error_message=f"CustomInstrument with symbol: S)invalid_universe.{uuid} for user: {uuid} "
            f"not found!. Requested ric: invalid_universe. Requested fields: None",
        )



from datetime import datetime
from tests.integration.conftest import (
    get_dict_values_as_list,
)


def on_refresh(data, instrument_name, stream, triggered_events_list=None):
    current_time = datetime.now().time()
    if triggered_events_list is not None:
        triggered_events_list.append("Refresh")
    print(f"{current_time} - Refresh received for {instrument_name} : {data}")


def on_updated(data, instrument_name, stream, triggered_events_list=None):
    current_time = datetime.now().time()
    if triggered_events_list is not None:
        triggered_events_list.append("Update")
    print(f"{current_time} - Update received for {instrument_name} : {data}")


def on_status(status, instrument_name, stream, triggered_events_list=None):
    current_time = datetime.now().time()
    if triggered_events_list is not None:
        triggered_events_list.append("Status")
    print(f"{current_time} - Status received for {instrument_name} : {status}")


def on_complete(stream, triggered_events_list=None):
    current_time = datetime.now().time()
    if triggered_events_list is not None:
        triggered_events_list.append("Complete")
    print(f"{current_time} - Stream is complete. Full snapshot:")
    print(stream.get_snapshot())


def on_error(status, instrument_name, stream, triggered_events_list=None):
    current_time = datetime.now().time()
    if triggered_events_list is not None:
        triggered_events_list.append("Error")
    print(f"{current_time} - Stream {stream} - Received error : {status}")


def add_callbacks_for_universe_stream(
    stream,
    triggered_events=None,
):
    stream.on_refresh(
        lambda data, instrument_name, stream_item: on_refresh(
            data, instrument_name, stream_item, triggered_events
        )
    )
    stream.on_update(
        lambda data, instrument_name, stream_item: on_updated(
            data, instrument_name, stream_item, triggered_events
        )
    )
    stream.on_complete(lambda stream_item: on_complete(stream_item, triggered_events))
    stream.on_status(
        lambda status, instrument_name, stream_item: on_status(
            status, instrument_name, stream_item, triggered_events
        )
    )
    stream.on_error(
        lambda status, instrument_name, stream_item: on_error(
            status, instrument_name, stream_item, triggered_events
        )
    )

    return stream


def check_triggered_events(triggered_events, expected_events):
    for i, event_type in enumerate(expected_events):
        assert (
            triggered_events[i] == event_type
        ), f"Event {triggered_events[i]} is not expected in order"


def assert_stream_error(stream, expected_msg, universe):
    actual_message = stream._stream._stream_by_name[universe].status["State"]["Text"]
    assert expected_msg in actual_message, f"{actual_message}"


def check_stream_data(stream, expected_universe, expected_fields=None):
    stream.open()

    try:
        if isinstance(expected_universe, str):
            expected_universe = [expected_universe]

        if isinstance(expected_universe, list):
            stream_item = stream._stream._stream_by_name
            for universe in expected_universe:
                content_value = get_dict_values_as_list(stream_item[universe])[0]
                assert content_value is not None, f"Stream object data in None"

        if isinstance(expected_fields, list):
            if isinstance(expected_universe, str):
                expected_universe = [expected_universe]
            for universe in expected_universe:
                actual_fields = set(
                    stream.get_snapshot(
                        universe=universe, fields=expected_fields
                    ).columns
                )
                actual_fields.remove("Instrument")
                assert (
                    set(expected_fields) == actual_fields
                ), f"{expected_fields} != {actual_fields}"

        if isinstance(expected_fields, str):
            actual_fields = stream.get_snapshot(universe=expected_universe).columns
            assert (
                expected_fields in actual_fields
            ), f"{expected_fields} not in {actual_fields}"

        if expected_fields is None:
            actual_fields = stream.get_snapshot(universe=expected_universe).columns
            assert (
                "PROD_PERM" in actual_fields or "TRDPRC_1" in actual_fields
            ), f"Fields does not exist in {actual_fields}"

    except KeyError as err:
        raise AssertionError(
            f"The instrument with universe {expected_universe} does not exist",
            f" \n The error appeared: {str(err)}",
        )

    stream.close()



from typing import Optional

from ._tools import _convert_date_time
from .._data._data_provider_layer import DataProviderLayer
from .._data._data_type import DataType
from ..._tools import validate_types


class Definition(DataProviderLayer):
    """
    Definition object.

    Parameters
    __________
        bucket : str
            The name of the bucket for file-sets to be searched.
            Only exactly matched results are returned.
        name : str, optional
            The name of the file-set. Only exactly matched results are returned.
        attributes : dict, optional
            Dict of publisher-defined key-value attributes.
        package_id : str, optional
        status : str, optional
             Filter file-set by status (Ready/Pending).
        available_from : str, optional
            Filter results by when the file was made available.
            The date/time format is YYYY-MM-DDTHH:mm:ssZ.(e.g. 2020-01-21T04:19:01Z)
        available_to : str, optional
            Filter results by when the file was made available.
            The date/time format is YYYY-MM-DDTHH:mm:ssZ.(e.g. 2020-01-21T04:19:01Z)
        content_from : str, optional
            Filter results by the age of the content within the file.
            The date/time format is YYYY-MM-DDTHH:mm:ssZ.(e.g. 2020-01-21T04:19:01Z)
        content_to : str, optional
            Filter results by the age of the content within the file.
            The date/time format is YYYY-MM-DDTHH:mm:ssZ.(e.g. 2020-01-21T04:19:01Z)
        created_since : str, optional
            Filter results by the date/time the file was created.
            The date/time format is YYYY-MM-DDTHH:mm:ssZ.(e.g. 2020-01-21T04:19:01Z)
        modified_since : str, optional
            Filter results by the date/time the file was updated.
            The date/time format is YYYY-MM-DDTHH:mm:ssZ. (e.g. 2020-01-21T04:19:01Z)
        skip_token : str, optional
            Skip token is only used if a previous operation returned a partial result.
            If a previous response contains a nextLink element, the value of the nextLink
            element will include a skip token parameter that specifies a starting point to
            use for subsequent calls.
        page_size : int, optional
            Number of fileset returned.
            By default 25.

    Methods
    -------
    get_data(session=session)
        Returns a response to the data platform
    get_data_async(session=None)
        Returns a response asynchronously to the data platform

    Examples
    --------
     >>> from refinitiv.data.delivery import cfs
     >>> definition = cfs.file_sets.Definition()
     >>> file_sets = definition.get_data()

     Using get_data_async
     >>> import asyncio
     >>> task = definition.get_data_async()
     >>> response = asyncio.run(task)
    """

    def __init__(
        self,
        bucket: str,
        name: Optional[str] = None,
        attributes: Optional[dict] = None,
        package_id: Optional[str] = None,
        status: Optional[str] = None,
        available_from: Optional[str] = None,
        available_to: Optional[str] = None,
        content_from: Optional[str] = None,
        content_to: Optional[str] = None,
        created_since: Optional[str] = None,
        modified_since: Optional[str] = None,
        skip_token: Optional[str] = None,
        page_size: int = 25,
    ):
        validate_types(page_size, [int], "page_size")
        created_since = _convert_date_time(created_since)
        modified_since = _convert_date_time(modified_since)
        available_from = _convert_date_time(available_from)
        available_to = _convert_date_time(available_to)
        content_from = _convert_date_time(content_from)
        content_to = _convert_date_time(content_to)
        if attributes is not None:
            attributes = ",".join(
                [f"{key}:{value}" for key, value in attributes.items()]
            )
        super().__init__(
            data_type=DataType.CFS_FILE_SETS,
            bucket=bucket,
            name=name,
            attributes=attributes,
            package_id=package_id,
            status=status,
            available_from=available_from,
            available_to=available_to,
            content_from=content_from,
            content_to=content_to,
            created_since=created_since,
            modified_since=modified_since,
            skip_token=skip_token,
            page_size=page_size,
        )



import os
from unittest.mock import MagicMock, patch

import pytest
from httpx import Client

from refinitiv.data import _configure as configure, OpenState
from refinitiv.data._core.session import PlatformSession, GrantPassword
from refinitiv.data.delivery._data._request import Request
from refinitiv.data.session import platform


# --------------------------------------------------------------------------------------
# self._auto_reconnect
# --------------------------------------------------------------------------------------


def test__auto_reconnect_property():
    # given
    expected_value = True
    definition = platform.Definition(
        app_key="app_key",
        grant=platform.GrantPassword(username="user", password="password"),
    )
    session = definition.get_session()

    # when
    testing_value = session.stream_auto_reconnection

    # then
    assert testing_value == expected_value


# --------------------------------------------------------------------------------------
# self._server_mode
# --------------------------------------------------------------------------------------


def test__server_mode_property():
    # given
    expected_value = False
    definition = platform.Definition(
        app_key="app_key",
        grant=platform.GrantPassword(username="user", password="password"),
    )
    session = definition.get_session()

    # when
    testing_value = session.server_mode

    # then
    assert testing_value == expected_value


# --------------------------------------------------------------------------------------
# self._base_url
# --------------------------------------------------------------------------------------


def test__base_url_property():
    # given
    expected_value = "https://api.refinitiv.com"
    definition = platform.Definition(
        app_key="app_key",
        grant=platform.GrantPassword(username="user", password="password"),
    )
    session = definition.get_session()

    # when
    testing_value = session._get_rdp_url_root()

    # then
    assert testing_value == expected_value


# --------------------------------------------------------------------------------------
# self._auth_url and self._auth_token
# --------------------------------------------------------------------------------------


def test_authentication_token_endpoint_url():
    # given
    expected_value = "https://api.refinitiv.com/auth/oauth2/v1/token"
    definition = platform.Definition(
        app_key="app_key",
        grant=platform.GrantPassword(username="user", password="password"),
    )
    session = definition.get_session()

    # when
    testing_value = session.authentication_token_endpoint_url

    # then
    assert testing_value == expected_value


def test__get_auth_token_uri():
    # given
    expected_value = "https://api.refinitiv.com/auth/oauth2/v1/token"
    definition = platform.Definition(
        app_key="app_key",
        grant=platform.GrantPassword(username="user", password="password"),
    )
    session = definition.get_session()

    # when
    testing_value = session._get_auth_token_uri()

    # then
    assert testing_value == expected_value


# --------------------------------------------------------------------------------------
# self._realtime_distribution_system_url
# --------------------------------------------------------------------------------------


def test__realtime_distribution_system_url_property():
    # given
    definition = platform.Definition(
        app_key="app_key",
        grant=platform.GrantPassword(username="user", password="password"),
    )
    session = definition.get_session()

    # when
    testing_value = session._deployed_platform_host

    # then
    assert testing_value is None


# --------------------------------------------------------------------------------------
# self._auth_authorize
# --------------------------------------------------------------------------------------


def test__auth_authorize_property():
    # given
    expected_value = "/authorize"
    definition = platform.Definition(
        app_key="app_key",
        grant=platform.GrantPassword(username="user", password="password"),
    )
    session = definition.get_session()

    # when
    testing_value = session._auth_authorize

    # then
    assert testing_value == expected_value


def test_session_name_is_set_right():
    session = PlatformSession(
        app_key="app_key",
        name="test",
        grant=platform.GrantPassword(username="user", password="password"),
    )
    assert session.name == "test"


def test_session_session_name_is_removed():
    session = PlatformSession(
        app_key="app_key",
        name="test",
        grant=platform.GrantPassword(username="user", password="password"),
    )
    with pytest.raises(AttributeError):
        session.session_name


def test_platform_session_has_config_snapshot_for_http_request_timeout_secs():
    # given
    session_a = PlatformSession(
        app_key="app_key",
        name="test",
        grant=platform.GrantPassword(username="user", password="password"),
    )

    # when
    config = configure.get_config()
    config["http.request-timeout"] = 1000
    session_b = PlatformSession(
        app_key="app_key",
        name="test",
        grant=platform.GrantPassword(username="user", password="password"),
    )

    # then
    assert session_a.http_request_timeout_secs != session_b.http_request_timeout_secs


def get_api_config(
    api_name,
    session,
):
    config = session.config
    api_config = config.get(f"apis.{api_name}")
    return api_config


def test_platform_session_has_config_snapshot_for_apis():
    # given
    session_a = PlatformSession(
        app_key="app_key",
        name="test",
        grant=platform.GrantPassword(username="user", password="password"),
    )

    # when
    config = configure.get_config()
    config["apis.data"] = {"changed key": "changed value"}
    session_b = PlatformSession(
        app_key="app_key",
        name="test",
        grant=platform.GrantPassword(username="user", password="password"),
    )

    # then
    api_config_a = get_api_config("data", session_a)
    api_config_b = get_api_config("data", session_b)
    assert api_config_a is not api_config_b


def test_platform_session_has_config_snapshot_for_get_streaming_websocket_endpoint_url():
    # given
    session_a = PlatformSession(
        app_key="app_key",
        name="test",
        grant=platform.GrantPassword(username="user", password="password"),
    )

    # when
    config = configure.get_config()
    config["apis.data"] = {"changed key": "changed value"}
    session_b = PlatformSession(
        app_key="app_key",
        name="test",
        grant=platform.GrantPassword(username="user", password="password"),
    )

    # then
    api_config_a = get_api_config("data", session_a)
    api_config_b = get_api_config("data", session_b)
    assert api_config_a is not api_config_b


def test_platform_session_has_config_snapshot_for_get_streaming_discovery_endpoint_url():
    # given
    session_a = PlatformSession(
        app_key="app_key",
        name="test",
        grant=platform.GrantPassword(username="user", password="password"),
    )

    # when
    config = configure.get_config()
    config["apis.data"] = {"changed key": "changed value"}
    session_b = PlatformSession(
        app_key="app_key",
        name="test",
        grant=platform.GrantPassword(username="user", password="password"),
    )

    # then
    api_config_a = get_api_config("data", session_a)
    api_config_b = get_api_config("data", session_b)
    assert api_config_a is not api_config_b


def test_platform_session_opens_and_closes_with_context_manager():
    def open_mock():
        return True

    # given
    grant = GrantPassword("username", "password")
    session = PlatformSession(app_key="app_key", grant=grant)
    origin_open = session._connection.open
    session._connection.open = open_mock

    # when
    with session:
        assert session.open_state is OpenState.Opened

    # then
    assert session.open_state is OpenState.Closed

    session._connection.open = origin_open


@pytest.mark.asyncio
async def test_platform_session_opens_and_closes_with_async_context_manager():
    def open_mock():
        return True

    # given
    grant = GrantPassword("username", "password")
    session = PlatformSession(app_key="app_key", grant=grant)
    origin_open = session._connection.open
    session._connection.open = open_mock

    # when
    async with session:
        assert session.open_state is OpenState.Opened

    # then
    assert session.open_state is OpenState.Closed

    session._connection.open = origin_open


@pytest.mark.parametrize(
    "input_value, expected_value",
    [
        (True, True),
        (False, False),
    ],
)
def test_signon_control(input_value, expected_value):
    # when
    session = PlatformSession(app_key="app_key", signon_control=input_value)

    # then
    assert session._take_signon_control is expected_value


def test_signon_control_default_value():
    # when
    session = PlatformSession(app_key="app_key")

    # then
    assert session._take_signon_control is True



from typing import TYPE_CHECKING, Any

from .._arg_enums import main_constituent_asset_class_arg_parser, risk_type_arg_parser
from .._base_data_class import BaseData
from ......delivery._data._endpoint_data import EndpointData

if TYPE_CHECKING:
    from ......_types import OptStr, OptBool
    from ...._curves._cross_currency_curves._types import (
        OptMainConstituentAssetClass,
        OptRiskType,
    )


curve_info_camel_to_snake = {
    "creationDateTime": "creation_date_time",
    "creationUserId": "creation_user_id",
    "updateDateTime": "update_date_time",
    "updateUserId": "update_user_id",
    "version": "version",
}

curve_definition_camel_to_snake = {
    "mainConstituentAssetClass": "main_constituent_asset_class",
    "riskType": "risk_type",
    "baseCurrency": "base_currency",
    "baseIndexName": "base_index_name",
    "definitionExpiryDate": "definition_expiry_date",
    "firstHistoricalAvailabilityDate": "first_historical_availability_date",
    "id": "id",
    "isFallbackForFxCurveDefinition": "is_fallback_for_fx_curve_definition",
    "isNonDeliverable": "is_non_deliverable",
    "name": "name",
    "quotedCurrency": "quoted_currency",
    "quotedIndexName": "quoted_index_name",
    "source": "source",
}


def convert_camel_to_snake(mapper: dict, data: dict) -> dict:
    return {mapper.get(name, name): value for name, value in data.items()}


class CurveInfo(BaseData):
    """
    Creates the Cross Currency curve definition with the definition provided.

    Parameters
    ----------
    creation_date_time : str, optional

    creation_user_id : str, optional

    update_date_time : str, optional

    update_user_id : str, optional

    version : str, optional

    """

    def __init__(
        self,
        creation_date_time: "OptStr" = None,
        creation_user_id: "OptStr" = None,
        update_date_time: "OptStr" = None,
        update_user_id: "OptStr" = None,
        version: "OptStr" = None,
        **kwargs,
    ):
        self.creation_date_time = creation_date_time
        self.creation_user_id = creation_user_id
        self.update_date_time = update_date_time
        self.update_user_id = update_user_id
        self.version = version
        super().__init__(**kwargs)


class CurveDefinition(BaseData):
    """
    Creates the Cross Currency curve definition with the definition provided.

    Parameters
    ----------
    main_constituent_asset_class : MainConstituentAssetClass, optional
        The asset class used to generate the zero coupon curve. the possible values are:
        * fxforward   * swap
    risk_type : RiskType, optional
        The risk type to which the generated cross currency curve is sensitive. the
        possible value is:   * 'crosscurrency'
    base_currency : str, optional
        The base currency in the fxcross currency pair. it is expressed in iso 4217
        alphabetical format (e.g., 'eur').
    base_index_name : str, optional
        The name of the floating rate index (e.g., 'estr') applied to the base currency.
    definition_expiry_date : str, optional
        The date after which curvedefinitions can not be used. the value is expressed in
        iso 8601 format: yyyy-mm-dd (e.g., '2021-01-01').
    first_historical_availability_date : str, optional
        The date starting from which cross currency curve definition can be used. the
        value is expressed in iso 8601 format: yyyy-mm-dd (e.g., '2021-01-01').
    id : str, optional
        The identifier of the cross currency curve.
    is_fallback_for_fx_curve_definition : bool, optional
        The indicator used to define the fallback logic for the fx curve definition. the
        possible values are:   * true: there's a fallback logic to use cross currency
        curve definition for fx curve definition,   * false: there's no fallback logic
        to use cross currency curve definition for fx curve definition.
    is_non_deliverable : bool, optional
        An indicator whether the instrument is non-deliverable:   * true: the instrument
        is non-deliverable,   * false: the instrument is not non-deliverable. the
        property can be used to retrieve cross currency definition for the adjusted
        interest rate curve.
    name : str, optional
        The fxcross currency pair applied to the reference or pivot currency. it is
        expressed in iso 4217 alphabetical format (e.g., 'eur usd fxcross').
    quoted_currency : str, optional
        The quoted currency in the fxcross currency pair. it is expressed in iso 4217
        alphabetical format (e.g., 'usd').
    quoted_index_name : str, optional
        The name of the floating rate index (e.g., 'sofr') applied to the quoted
        currency.
    source : str, optional
        A user-defined string that is provided by the creator of a curve. curves created
        by refinitiv have the 'refinitiv' source.
    """

    def __init__(
        self,
        main_constituent_asset_class: "OptMainConstituentAssetClass" = None,
        risk_type: "OptRiskType" = None,
        base_currency: "OptStr" = None,
        base_index_name: "OptStr" = None,
        definition_expiry_date: "OptStr" = None,
        first_historical_availability_date: "OptStr" = None,
        id: "OptStr" = None,
        is_fallback_for_fx_curve_definition: "OptBool" = None,
        is_non_deliverable: "OptBool" = None,
        name: "OptStr" = None,
        quoted_currency: "OptStr" = None,
        quoted_index_name: "OptStr" = None,
        source: "OptStr" = None,
        **kwargs,
    ):
        self.main_constituent_asset_class = (
            main_constituent_asset_class_arg_parser.get_enum(
                main_constituent_asset_class
            )
        )
        self.risk_type = risk_type_arg_parser.get_enum(risk_type)
        self.base_currency = base_currency
        self.base_index_name = base_index_name
        self.definition_expiry_date = definition_expiry_date
        self.first_historical_availability_date = first_historical_availability_date
        self.id = id
        self.is_fallback_for_fx_curve_definition = is_fallback_for_fx_curve_definition
        self.is_non_deliverable = is_non_deliverable
        self.name = name
        self.quoted_currency = quoted_currency
        self.quoted_index_name = quoted_index_name
        self.source = source
        super().__init__(**kwargs)


class CurveDefinitionData(EndpointData):
    def __init__(
        self,
        raw: Any,
        **kwargs,
    ):
        EndpointData.__init__(self, raw, **kwargs)
        self._curve_definition = None
        self._curve_info = None

    @property
    def curve_definition(self):
        if self._curve_definition is None:
            curve_definition = self.raw.get("curveDefinition")
            if curve_definition:
                curve_definition = convert_camel_to_snake(
                    curve_definition_camel_to_snake, curve_definition
                )
            self._curve_definition = CurveDefinition(**curve_definition)
        return self._curve_definition

    @property
    def curve_info(self):
        if self._curve_info is None:
            curve_info = self.raw.get("curveInfo")
            if curve_info:
                curve_info = convert_camel_to_snake(
                    curve_info_camel_to_snake, curve_info
                )
            self._curve_info = CurveInfo(**curve_info)
        return self._curve_info



import pytest

import refinitiv.data as rd


@pytest.fixture(scope="function", autouse=False)
def enabled_log():
    rd.get_config().set_param(param=f"logs.transports.console.enabled", value=True)
    rd.get_config().set_param(param=f"logs.transports.file.enabled", value=True)



from typing import TYPE_CHECKING, List, Set

if TYPE_CHECKING:
    from .delivery._data._data_provider import Response


class RDError(Exception):
    """Base class for exceptions in this refinitiv-data module.

    Parameters
    ----------
    code: int
        error code for this exception
    message: str
        error description for this exception
    """

    response: "Response"

    def __init__(self, code: int, message: str):
        self.code = code
        self.message = message

    def __str__(self) -> str:
        return f"Error code {self.code} | {self.message}"


class SessionError(RDError):
    pass


class StreamingError(RDError):
    pass


class StreamConnectionError(RDError):
    pass


class PlatformSessionError(SessionError):
    pass


class DesktopSessionError(SessionError):
    pass


class ESGError(RDError):
    pass


class NewsHeadlinesError(RDError):
    pass


class EndpointError(RDError):
    pass


class StreamError(RDError):
    pass


class StreamingPricesError(RDError):
    pass


class ItemWasNotRequested(RDError):
    def __init__(self, item_type, not_requested, requested):
        item_type = f"{item_type}{'s' if len(not_requested) > 1 else ''}"
        super().__init__(
            -1, f"{item_type} {not_requested} was not requested : {requested}"
        )


class EnvError(RDError):
    pass


class BondPricingError(RDError):
    pass


class FinancialContractsError(RDError):
    pass


class RequiredError(RDError):
    pass


class ConfigurationError(RDError):
    pass


class ScopeError(RDError):
    def __init__(
        self,
        required_scope: List[Set[str]],
        available_scope: Set[str],
        key: str,
        method: str,
    ):
        missing_scopes = (scope - available_scope for scope in required_scope)
        self.required_scope = required_scope
        self.available_scope = available_scope
        self.key = key
        self.method = method
        self.missing_scopes = missing_scopes
        super().__init__(
            -1,
            message=f"Insufficient scope for key={key}, method={method} failed.\n"
            f"Required scope: {' OR '.join(map(str, required_scope))}\n"
            f"Token scope: {available_scope}\n"
            f"Missing scopes: {' OR '.join(map(str, missing_scopes))}",
        )



import allure
import pytest

import refinitiv.data as rd
from refinitiv.data._fin_coder_layer import Format
from refinitiv.data.errors import RDError
from tests.integration.access.news.conftest import check_story_response_format

ERROR_MESSAGE_PATTERN = {
    "PLATFORM": "Error code 404 | Error while calling the NEP backend: Story not found",
    "DESKTOP": "Error code 404 | Error while calling the NEP backend: Story not found",
}


@allure.suite("FinCoder Layer")
@allure.feature("FinCoder - News Story")
@allure.severity(allure.severity_level.CRITICAL)
@pytest.mark.underlying_content("news")
class TestGetStory:
    @allure.title("Get News Story with opened session and valid story ID")
    @pytest.mark.caseid("")
    @pytest.mark.parametrize(
        "response_format",
        [(Format.HTML), (Format.TEXT)],
    )
    def test_get_news_story_with_valid_story_id(self, open_session, response_format):
        story_keywords = ["NASA", "SpaceX", "mission"]

        def get_format(session, story_format):
            if session.type.name == "PLATFORM" and story_format == Format.HTML:
                return "Html"
            elif session.type.name == "DESKTOP" and story_format == Format.TEXT:
                return "Text"

        _format = get_format(open_session, response_format) or response_format
        response = rd.news.get_story(
            story_id="urn:newsml:reuters.com:20211125:nNRAhybonl:1",
            format=_format,
        )

        assert response, f"Response is empty - {response}"
        assert isinstance(response, str), f"Story has type {type(response)}"
        assert any(
            key in response for key in story_keywords
        ), "The unexpected story is responded"
        check_story_response_format(response, response_format)

    @allure.title("Get News Story using opened session with the invalid storyId ")
    @pytest.mark.caseid("")
    @pytest.mark.parametrize("id", ["invalid"])
    def test_get_news_story_with_invalid_story_id(self, open_session, id):
        underlying_mark = open_session.type.name
        expected_error_message = ERROR_MESSAGE_PATTERN[underlying_mark]
        with pytest.raises(RDError) as error:
            rd.news.get_story(id)
        assert str(error.value) == expected_error_message



from unittest.mock import patch, MagicMock

import pytest

from refinitiv.data import open_pricing_stream
from refinitiv.data._core.session import set_default
from refinitiv.data._fin_coder_layer._mixed_streams import MixedStreams
from tests.unit.conftest import StubSession


def test_session_is_close():
    # given
    error_message = "Session is not opened. Can't send any request"

    session = StubSession()
    set_default(session)

    # when
    with pytest.raises(ValueError, match=error_message):
        open_pricing_stream(...)


@patch("refinitiv.data._fin_coder_layer.get_stream.Stream")
def test_open_pricing_stream(mock_stream):
    session = StubSession(is_open=True)
    set_default(session)

    callback = lambda *args, **kwargs: print(args, kwargs)

    stream = open_pricing_stream(
        universe=["EUR=", "USD="], fields=["BID", "ASK", "OPEN_PRC"], on_data=callback
    )

    mock_stream.assert_called_once()
    attr_list = ["get_snapshot", "open", "close", "recorder"]
    for attr in attr_list:
        assert hasattr(stream, attr)


def test_mixed_streams():
    def add_to_list(_list: list, elem):
        _list.append(elem)
        return MagicMock()

    pricing = []
    custom_insts = []
    mixed_streams = MixedStreams(
        universe=["S)Test1", "Test2", "S)Test3", "S)Test4", "Test5"],
        session=StubSession(),
    )
    mixed_streams._get_pricing_stream = lambda a: add_to_list(pricing, a)
    mixed_streams._get_custom_instruments_stream = lambda a: add_to_list(
        custom_insts, a
    )
    mixed_streams._stream_by_name

    assert pricing == ["Test2", "Test5"]
    assert custom_insts == ["S)Test1", "S)Test3", "S)Test4"]



from refinitiv.data.content.ipa._curves._cross_currency_curves._triangulate_definitions import (
    RequestItem,
)


def test_definitions_triangulate_request_item():
    # given
    expected_dict = {
        "baseCurrency": "string",
        "baseIndexName": "string",
        "curveTag": "string",
        "quotedCurrency": "string",
        "quotedIndexName": "string",
        "valuationDate": "string",
    }

    # when
    testing_obj = RequestItem(
        base_currency="string",
        base_index_name="string",
        curve_tag="string",
        quoted_currency="string",
        quoted_index_name="string",
        valuation_date="string",
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.base_currency == "string"
    assert testing_obj.base_index_name == "string"
    assert testing_obj.curve_tag == "string"
    assert testing_obj.quoted_currency == "string"
    assert testing_obj.quoted_index_name == "string"
    assert testing_obj.valuation_date == "string"



# coding: utf8

__all__ = ["DayWeight"]

from .._object_definition import ObjectDefinition


class DayWeight(ObjectDefinition):
    def __init__(self, date=None, weight=None):
        super().__init__()
        self.date = date
        self.weight = weight

    @property
    def date(self):
        """
        :return: str
        """
        return self._get_parameter("date")

    @date.setter
    def date(self, value):
        self._set_parameter("date", value)

    @property
    def weight(self):
        """
        :return: float
        """
        return self._get_parameter("weight")

    @weight.setter
    def weight(self, value):
        self._set_parameter("weight", value)



from typing import List

from ._top_news_headline import TopNewsHeadline
from ..._content_data import Data
from ..._content_data_provider import ContentDataProvider
from ..._content_response_factory import ContentResponseFactory
from ...._tools import cached_property, ParamItem
from ....delivery._data._request_factory import RequestFactory


class TopNewsData(Data):
    @cached_property
    def headlines(self) -> "List[TopNewsHeadline]":
        return [
            TopNewsHeadline.from_dict(headline_data)
            for headline_data in self.raw.get("data", [])
        ]


query_params = [ParamItem("revision_id", "revisionId")]


class TopNewsRequestFactory(RequestFactory):
    def get_path_parameters(self, session=None, *, top_news_id=None, **kwargs):
        return {"topNewsId": top_news_id}

    def get_url(self, *args, **kwargs):
        return f"{super().get_url(*args, **kwargs)}/{{topNewsId}}"

    @property
    def query_params_config(self):
        return query_params


news_top_news_data_provider = ContentDataProvider(
    request=TopNewsRequestFactory(),
    response=ContentResponseFactory(data_class=TopNewsData),
)
